# Copyright 2014-2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime, timedelta
import json
import tempfile
import unittest
from mock import call, MagicMock, patch, sentinel

from charmhelpers import coordinator
from charmhelpers.core import hookenv


class TestCoordinator(unittest.TestCase):

    def setUp(self):
        del hookenv._atstart[:]
        del hookenv._atexit[:]
        hookenv.cache.clear()
        coordinator.Singleton._instances.clear()

        def install(patch):
            patch.start()
            self.addCleanup(patch.stop)

        install(patch.object(hookenv, 'local_unit', return_value='foo/1'))
        install(patch.object(hookenv, 'is_leader', return_value=False))
        install(patch.object(hookenv, 'metadata',
                             return_value={'peers': {'cluster': None}}))
        install(patch.object(hookenv, 'log'))

        # Ensure _timestamp always increases.
        install(patch.object(coordinator, '_utcnow',
                             side_effect=self._utcnow))

    _last_utcnow = datetime(2015, 1, 1, 00, 00)

    def _utcnow(self, ts=coordinator._timestamp):
        self._last_utcnow += timedelta(minutes=1)
        return self._last_utcnow

    def test_is_singleton(self):
        # BaseCoordinator and subclasses are singletons. Placing this
        # burden on charm authors is impractical, particularly if
        # libraries start wanting to use coordinator instances.
        # With singletons, we don't need to worry about sharing state
        # between instances or have them stomping on each other when they
        # need to serialize their state.
        self.assertTrue(coordinator.BaseCoordinator()
                        is coordinator.BaseCoordinator())
        self.assertTrue(coordinator.Serial() is coordinator.Serial())
        self.assertFalse(coordinator.BaseCoordinator() is coordinator.Serial())

    @patch.object(hookenv, 'atstart')
    def test_implicit_initialize_and_handle(self, atstart):
        # When you construct a BaseCoordinator(), its initialize() and
        # handle() method are invoked automatically every hook. This
        # is done using hookenv.atstart
        c = coordinator.BaseCoordinator()
        atstart.assert_has_calls([call(c.initialize), call(c.handle)])

    @patch.object(hookenv, 'has_juju_version', return_value=False)
    def test_initialize_enforces_juju_version(self, has_juju_version):
        c = coordinator.BaseCoordinator()
        with self.assertRaises(AssertionError):
            c.initialize()
        has_juju_version.assert_called_once_with('1.23')

    @patch.object(hookenv, 'atexit')
    @patch.object(hookenv, 'has_juju_version', return_value=True)
    @patch.object(hookenv, 'relation_ids')
    def test_initialize(self, relation_ids, ver, atexit):
        # First initialization are done before there is a peer relation.
        relation_ids.return_value = []
        c = coordinator.BaseCoordinator()

        with patch.object(c, '_load_state') as _load_state, \
                patch.object(c, '_emit_state') as _emit_state:  # IGNORE: E127
            c.initialize()
            _load_state.assert_called_once_with()
            _emit_state.assert_called_once_with()

        self.assertEqual(c.relname, 'cluster')
        self.assertIsNone(c.relid)
        relation_ids.assert_called_once_with('cluster')

        # Methods installed to save state and release locks if the
        # hook is successful.
        atexit.assert_has_calls([call(c._save_state),
                                 call(c._release_granted)])

        # If we have a peer relation, the id is stored.
        relation_ids.return_value = ['cluster:1']
        c = coordinator.BaseCoordinator()
        with patch.object(c, '_load_state'), patch.object(c, '_emit_state'):
            c.initialize()
        self.assertEqual(c.relid, 'cluster:1')

        # If we are already initialized, nothing happens.
        c.grants = {}
        c.requests = {}
        c.initialize()

    def test_acquire(self):
        c = coordinator.BaseCoordinator()
        lock = 'mylock'
        c.grants = {}
        c.requests = {hookenv.local_unit(): {}}

        # We are not the leader, so first acquire will return False.
        self.assertFalse(c.acquire(lock))

        # But the request is in the queue.
        self.assertTrue(c.requested(lock))
        ts = c.request_timestamp(lock)

        # A further attempts at acquiring the lock do nothing,
        # and the timestamp of the request remains unchanged.
        self.assertFalse(c.acquire(lock))
        self.assertEqual(ts, c.request_timestamp(lock))

        # Once the leader has granted the lock, acquire returns True.
        with patch.object(c, 'granted') as granted:
            granted.return_value = True
            self.assertTrue(c.acquire(lock))
            granted.assert_called_once_with(lock)

    def test_acquire_leader(self):
        # When acquire() is called by the leader, it needs
        # to make a grant decision immediately. It can't defer
        # making the decision until a future hook, as no future
        # hooks will be triggered.
        hookenv.is_leader.return_value = True
        c = coordinator.Serial()  # Not Base. Test hooks into default_grant.
        lock = 'mylock'
        unit = hookenv.local_unit()
        c.grants = {}
        c.requests = {unit: {}}
        with patch.object(c, 'default_grant') as default_grant:
            default_grant.side_effect = iter([False, True])

            self.assertFalse(c.acquire(lock))
            ts = c.request_timestamp(lock)

            self.assertTrue(c.acquire(lock))
            self.assertEqual(ts, c.request_timestamp(lock))

            # If it it granted, the leader doesn't make a decision again.
            self.assertTrue(c.acquire(lock))
            self.assertEqual(ts, c.request_timestamp(lock))

            self.assertEqual(default_grant.call_count, 2)

    def test_granted(self):
        c = coordinator.BaseCoordinator()
        unit = hookenv.local_unit()
        lock = 'mylock'
        ts = coordinator._timestamp()
        c.grants = {}

        # Unit makes a request, but it isn't granted
        c.requests = {unit: {lock: ts}}
        self.assertFalse(c.granted(lock))

        # Once the leader has granted the request, all good.
        # It does this by mirroring the request timestamp.
        c.grants = {unit: {lock: ts}}
        self.assertTrue(c.granted(lock))

        # The unit releases the lock by removing the request.
        c.requests = {unit: {}}
        self.assertFalse(c.granted(lock))

        # If the unit makes a new request before the leader
        # has had a chance to do its housekeeping, the timestamps
        # do not match and the lock not considered granted.
        ts = coordinator._timestamp()
        c.requests = {unit: {lock: ts}}
        self.assertFalse(c.granted(lock))

        # Until the leader gets around to its duties.
        c.grants = {unit: {lock: ts}}
        self.assertTrue(c.granted(lock))

    def test_requested(self):
        c = coordinator.BaseCoordinator()
        lock = 'mylock'
        c.requests = {hookenv.local_unit(): {}}
        c.grants = {}

        self.assertFalse(c.requested(lock))
        c.acquire(lock)
        self.assertTrue(c.requested(lock))

    def test_request_timestamp(self):
        c = coordinator.BaseCoordinator()
        lock = 'mylock'
        unit = hookenv.local_unit()

        c.requests = {unit: {}}
        c.grants = {}
        self.assertIsNone(c.request_timestamp(lock))

        now = datetime.utcnow()
        fmt = coordinator._timestamp_format
        c.requests = {hookenv.local_unit(): {lock: now.strftime(fmt)}}

        self.assertEqual(c.request_timestamp(lock), now)

    def test_handle_not_leader(self):
        c = coordinator.BaseCoordinator()
        # If we are not the leader, handle does nothing. We know this,
        # because without mocks or initialization it would otherwise crash.
        c.handle()

    def test_handle(self):
        hookenv.is_leader.return_value = True
        lock = 'mylock'
        c = coordinator.BaseCoordinator()
        c.relid = 'cluster:1'

        ts = coordinator._timestamp
        ts1, ts2, ts3 = ts(), ts(), ts()

        # Grant one of these requests.
        requests = {'foo/1': {lock: ts1},
                    'foo/2': {lock: ts2},
                    'foo/3': {lock: ts3}}
        c.requests = requests.copy()
        # Because the existing grant should be released.
        c.grants = {'foo/2': {lock: ts()}}  # No request, release.

        with patch.object(c, 'grant') as grant:
            c.handle()

            # The requests are unchanged. This is normally state on the
            # peer relation, and only the units themselves can change it.
            self.assertDictEqual(requests, c.requests)

            # The grant without a corresponding requests was released.
            self.assertDictEqual({'foo/2': {}}, c.grants)

            # A potential grant was made for each of the outstanding requests.
            grant.assert_has_calls([call(lock, 'foo/1'),
                                    call(lock, 'foo/2'),
                                    call(lock, 'foo/3')], any_order=True)

    def test_grant_not_leader(self):
        c = coordinator.BaseCoordinator()
        c.grant(sentinel.whatever, sentinel.whatever)  # Nothing happens.

    def test_grant(self):
        hookenv.is_leader.return_value = True
        c = coordinator.BaseCoordinator()
        c.default_grant = MagicMock()
        c.grant_other = MagicMock()

        ts = coordinator._timestamp
        ts1, ts2 = ts(), ts()

        c.requests = {'foo/1': {'mylock': ts1, 'other': ts()},
                      'foo/2': {'mylock': ts2},
                      'foo/3': {'mylock': ts()}}
        grants = {'foo/1': {'mylock': ts1}}
        c.grants = grants.copy()

        # foo/1 already has a granted mylock, so returns True.
        self.assertTrue(c.grant('mylock', 'foo/1'))

        # foo/2 does not have a granted mylock. default_grant will
        # be called to make a decision (no)
        c.default_grant.return_value = False
        self.assertFalse(c.grant('mylock', 'foo/2'))
        self.assertDictEqual(grants, c.grants)
        c.default_grant.assert_called_once_with('mylock', 'foo/2',
                                                set(['foo/1']),
                                                ['foo/2', 'foo/3'])
        c.default_grant.reset_mock()

        # Lets say yes.
        c.default_grant.return_value = True
        self.assertTrue(c.grant('mylock', 'foo/2'))
        grants = {'foo/1': {'mylock': ts1}, 'foo/2': {'mylock': ts2}}
        self.assertDictEqual(grants, c.grants)
        c.default_grant.assert_called_once_with('mylock', 'foo/2',
                                                set(['foo/1']),
                                                ['foo/2', 'foo/3'])

        # The other lock has custom logic, in the form of the overridden
        # grant_other method.
        c.grant_other.return_value = False
        self.assertFalse(c.grant('other', 'foo/1'))
        c.grant_other.assert_called_once_with('other', 'foo/1',
                                              set(), ['foo/1'])

        # If there is no request, grant returns False
        c.grant_other.return_value = True
        self.assertFalse(c.grant('other', 'foo/2'))

    def test_released(self):
        c = coordinator.BaseCoordinator()
        with patch.object(c, 'msg') as msg:
            c.released('foo/2', 'mylock', coordinator._utcnow())
            expected = 'Leader released mylock from foo/2, held 0:01:00'
            msg.assert_called_once_with(expected)

    def test_require(self):
        c = coordinator.BaseCoordinator()
        c.acquire = MagicMock()
        c.granted = MagicMock()
        guard = MagicMock()

        wrapped = MagicMock()

        @c.require('mylock', guard)
        def func(*args, **kw):
            wrapped(*args, **kw)

        # If the lock is granted, the wrapped function is called.
        c.granted.return_value = True
        func(arg=True)
        wrapped.assert_called_once_with(arg=True)
        wrapped.reset_mock()

        # If the lock is not granted, and the guard returns False,
        # the lock is not acquired.
        c.acquire.return_value = False
        c.granted.return_value = False
        guard.return_value = False
        func()
        self.assertFalse(wrapped.called)
        self.assertFalse(c.acquire.called)

        # If the lock is not granted, and the guard returns True,
        # the lock is acquired. But the function still isn't called if
        # it cannot be acquired immediately.
        guard.return_value = True
        func()
        self.assertFalse(wrapped.called)
        c.acquire.assert_called_once_with('mylock')

        # Finally, if the lock is not granted, and the guard returns True,
        # and the lock acquired immediately, the function is called.
        c.acquire.return_value = True
        func(sentinel.arg)
        wrapped.assert_called_once_with(sentinel.arg)

    def test_msg(self):
        c = coordinator.BaseCoordinator()
        # Just a wrapper around hookenv.log
        c.msg('hi')
        hookenv.log.assert_called_once_with('coordinator.BaseCoordinator hi',
                                            level=hookenv.INFO)

    def test_name(self):
        # We use the class name in a few places to avoid conflicts.
        # We assume we won't be using multiple BaseCoordinator subclasses
        # with the same name at the same time.
        c = coordinator.BaseCoordinator()
        self.assertEqual(c._name(), 'BaseCoordinator')
        c = coordinator.Serial()
        self.assertEqual(c._name(), 'Serial')

    @patch.object(hookenv, 'leader_get')
    def test_load_state(self, leader_get):
        c = coordinator.BaseCoordinator()
        unit = hookenv.local_unit()

        # c.granted is just the leader_get decoded.
        leader_get.return_value = '{"json": true}'
        c._load_state()
        self.assertDictEqual(c.grants, {'json': True})

        # With no relid, there is no peer relation so request state
        # is pulled from a local stash.
        with patch.object(c, '_load_local_state') as loc_state:
            loc_state.return_value = {'local': True}
            c._load_state()
            self.assertDictEqual(c.requests, {unit: {'local': True}})

        # With a relid, request details are pulled from the peer relation.
        # If there is no data in the peer relation from the local unit,
        # we still pull it from the local stash as it means this is the
        # first time we have joined.
        c.relid = 'cluster:1'
        with patch.object(c, '_load_local_state') as loc_state, \
                patch.object(c, '_load_peer_state') as peer_state:
            loc_state.return_value = {'local': True}
            peer_state.return_value = {'foo/2': {'mylock': 'whatever'}}
            c._load_state()
            self.assertDictEqual(c.requests, {unit: {'local': True},
                                              'foo/2': {'mylock': 'whatever'}})

        # If there are local details in the peer relation, the local
        # stash is ignored.
        with patch.object(c, '_load_local_state') as loc_state, \
                patch.object(c, '_load_peer_state') as peer_state:
            loc_state.return_value = {'local': True}
            peer_state.return_value = {unit: {},
                                       'foo/2': {'mylock': 'whatever'}}
            c._load_state()
            self.assertDictEqual(c.requests, {unit: {},
                                              'foo/2': {'mylock': 'whatever'}})

    def test_emit_state(self):
        c = coordinator.BaseCoordinator()
        unit = hookenv.local_unit()
        c.requests = {unit: {'lock_a': sentinel.ts,
                             'lock_b': sentinel.ts,
                             'lock_c': sentinel.ts}}
        c.grants = {unit: {'lock_a': sentinel.ts,
                           'lock_b': sentinel.ts2}}
        with patch.object(c, 'msg') as msg:
            c._emit_state()
            msg.assert_has_calls([call('Granted lock_a'),
                                  call('Waiting on lock_b'),
                                  call('Waiting on lock_c')],
                                 any_order=True)

    @patch.object(hookenv, 'relation_set')
    @patch.object(hookenv, 'leader_set')
    def test_save_state(self, leader_set, relation_set):
        c = coordinator.BaseCoordinator()
        unit = hookenv.local_unit()
        c.grants = {'directdump': True}
        c.requests = {unit: 'data1', 'foo/2': 'data2'}

        # grants is dumped to leadership settings, if the unit is leader.
        with patch.object(c, '_save_local_state') as save_loc:
            c._save_state()
            self.assertFalse(leader_set.called)
            hookenv.is_leader.return_value = True
            c._save_state()
            leader_set.assert_called_once_with({c.key: '{"directdump": true}'})

        # If there is no relation id, the local units requests is dumped
        # to a local stash.
        with patch.object(c, '_save_local_state') as save_loc:
            c._save_state()
            save_loc.assert_called_once_with('data1')

        # If there is a relation id, the local units requests is dumped
        # to the peer relation.
        with patch.object(c, '_save_local_state') as save_loc:
            c.relid = 'cluster:1'
            c._save_state()
            self.assertFalse(save_loc.called)
            relation_set.assert_called_once_with(
                c.relid, relation_settings={c.key: '"data1"'})  # JSON encoded

    @patch.object(hookenv, 'relation_get')
    @patch.object(hookenv, 'related_units')
    def test_load_peer_state(self, related_units, relation_get):
        # Standard relation-get loops, decoding results from JSON.
        c = coordinator.BaseCoordinator()
        c.key = sentinel.key
        c.relid = sentinel.relid
        related_units.return_value = ['foo/2', 'foo/3']
        d = {'foo/1': {'foo/1': True},
             'foo/2': {'foo/2': True},
             'foo/3': {'foo/3': True}}

        def _get(key, unit, relid):
            assert key == sentinel.key
            assert relid == sentinel.relid
            return json.dumps(d[unit])
        relation_get.side_effect = _get

        self.assertDictEqual(c._load_peer_state(), d)

    def test_local_state_filename(self):
        c = coordinator.BaseCoordinator()
        self.assertEqual(c._local_state_filename(),
                         '.charmhelpers.coordinator.BaseCoordinator')

    def test_load_local_state(self):
        c = coordinator.BaseCoordinator()
        with tempfile.NamedTemporaryFile(mode='w') as f:
            with patch.object(c, '_local_state_filename') as fn:
                fn.return_value = f.name
                d = 'some data'
                json.dump(d, f)
                f.flush()
                d2 = c._load_local_state()
                self.assertEqual(d, d2)

    def test_save_local_state(self):
        c = coordinator.BaseCoordinator()
        with tempfile.NamedTemporaryFile(mode='r') as f:
            with patch.object(c, '_local_state_filename') as fn:
                fn.return_value = f.name
                c._save_local_state('some data')
                self.assertEqual(json.load(f), 'some data')

    def test_release_granted(self):
        c = coordinator.BaseCoordinator()
        unit = hookenv.local_unit()
        c.requests = {unit: {'lock1': sentinel.ts, 'lock2': sentinel.ts},
                      'foo/2': {'lock1': sentinel.ts}}
        c.grants = {unit: {'lock1': sentinel.ts},
                    'foo/2': {'lock1': sentinel.ts}}
        # The granted lock for the local unit is released.
        c._release_granted()
        self.assertDictEqual(c.requests, {unit: {'lock2': sentinel.ts},
                                          'foo/2': {'lock1': sentinel.ts}})

    def test_implicit_peer_relation_name(self):
        self.assertEqual(coordinator._implicit_peer_relation_name(),
                         'cluster')

    def test_default_grant(self):
        c = coordinator.Serial()
        # Lock not granted. First in the queue.
        self.assertTrue(c.default_grant(sentinel.lock, sentinel.u1,
                                        set(), [sentinel.u1, sentinel.u2]))

        # Lock not granted. Later in the queue.
        self.assertFalse(c.default_grant(sentinel.lock, sentinel.u1,
                                         set(), [sentinel.u2, sentinel.u1]))

        # Lock already granted
        self.assertFalse(c.default_grant(sentinel.lock, sentinel.u1,
                                         set([sentinel.u2]), [sentinel.u1]))



# Copyright 2019 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import configparser
import glob
import os.path
import subprocess

from charmhelpers.contrib.openstack.audits import (
    audit,
    AuditType,
    # filters
    is_audit_type,
    it_has_config,
)

from charmhelpers.core.hookenv import (
    cached,
)

"""
The Security Guide suggests a specific list of files inside the
config directory for the service having 640 specifically, but
by ensuring the containing directory is 750, only the owner can
write, and only the group can read files within the directory.

By  restricting access to the containing directory, we can more
effectively ensure that there is no accidental leakage if a new
file is added to the service without being added to the security
guide, and to this check.
"""
FILE_ASSERTIONS = {
    'barbican': {
        '/etc/barbican': {'group': 'barbican', 'mode': '750'},
    },
    'ceph-mon': {
        '/var/lib/charm/ceph-mon/ceph.conf':
            {'owner': 'root', 'group': 'root', 'mode': '644'},
        '/etc/ceph/ceph.client.admin.keyring':
            {'owner': 'ceph', 'group': 'ceph'},
        '/etc/ceph/rbdmap': {'mode': '644'},
        '/var/lib/ceph': {'owner': 'ceph', 'group': 'ceph', 'mode': '750'},
        '/var/lib/ceph/bootstrap-*/ceph.keyring':
            {'owner': 'ceph', 'group': 'ceph', 'mode': '600'}
    },
    'ceph-osd': {
        '/var/lib/charm/ceph-osd/ceph.conf':
            {'owner': 'ceph', 'group': 'ceph', 'mode': '644'},
        '/var/lib/ceph': {'owner': 'ceph', 'group': 'ceph', 'mode': '750'},
        '/var/lib/ceph/*': {'owner': 'ceph', 'group': 'ceph', 'mode': '755'},
        '/var/lib/ceph/bootstrap-*/ceph.keyring':
            {'owner': 'ceph', 'group': 'ceph', 'mode': '600'},
        '/var/lib/ceph/radosgw':
            {'owner': 'ceph', 'group': 'ceph', 'mode': '755'},
    },
    'cinder': {
        '/etc/cinder': {'group': 'cinder', 'mode': '750'},
    },
    'glance': {
        '/etc/glance': {'group': 'glance', 'mode': '750'},
    },
    'keystone': {
        '/etc/keystone':
            {'owner': 'keystone', 'group': 'keystone', 'mode': '750'},
    },
    'manilla': {
        '/etc/manila': {'group': 'manilla', 'mode': '750'},
    },
    'neutron-gateway': {
        '/etc/neutron': {'group': 'neutron', 'mode': '750'},
    },
    'neutron-api': {
        '/etc/neutron/': {'group': 'neutron', 'mode': '750'},
    },
    'nova-cloud-controller': {
        '/etc/nova': {'group': 'nova', 'mode': '750'},
    },
    'nova-compute': {
        '/etc/nova/': {'group': 'nova', 'mode': '750'},
    },
    'openstack-dashboard': {
        # From security guide
        '/etc/openstack-dashboard/local_settings.py':
            {'group': 'horizon', 'mode': '640'},
    },
}

Ownership = collections.namedtuple('Ownership', 'owner group mode')


@cached
def _stat(file):
    """
    Get the Ownership information from a file.

    :param file: The path to a file to stat
    :type file: str
    :returns: owner, group, and mode of the specified file
    :rtype: Ownership
    :raises subprocess.CalledProcessError: If the underlying stat fails
    """
    out = subprocess.check_output(
        ['stat', '-c', '%U %G %a', file]).decode('utf-8')
    return Ownership(*out.strip().split(' '))


@cached
def _config_ini(path):
    """
    Parse an ini file

    :param path: The path to a file to parse
    :type file: str
    :returns: Configuration contained in path
    :rtype: Dict
    """
    # When strict is enabled, duplicate options are not allowed in the
    # parsed INI; however, Oslo allows duplicate values. This change
    # causes us to ignore the duplicate values which is acceptable as
    # long as we don't validate any multi-value options
    conf = configparser.ConfigParser(strict=False)
    conf.read(path)
    return dict(conf)


def _validate_file_ownership(owner, group, file_name, optional=False):
    """
    Validate that a specified file is owned by `owner:group`.

    :param owner: Name of the owner
    :type owner: str
    :param group: Name of the group
    :type group: str
    :param file_name: Path to the file to verify
    :type file_name: str
    :param optional: Is this file optional,
                     ie: Should this test fail when it's missing
    :type optional: bool
    """
    try:
        ownership = _stat(file_name)
    except subprocess.CalledProcessError as e:
        print("Error reading file: {}".format(e))
        if not optional:
            assert False, "Specified file does not exist: {}".format(file_name)
    assert owner == ownership.owner, \
        "{} has an incorrect owner: {} should be {}".format(
            file_name, ownership.owner, owner)
    assert group == ownership.group, \
        "{} has an incorrect group: {} should be {}".format(
            file_name, ownership.group, group)
    print("Validate ownership of {}: PASS".format(file_name))


def _validate_file_mode(mode, file_name, optional=False):
    """
    Validate that a specified file has the specified permissions.

    :param mode: file mode that is desires
    :type owner: str
    :param file_name: Path to the file to verify
    :type file_name: str
    :param optional: Is this file optional,
                     ie: Should this test fail when it's missing
    :type optional: bool
    """
    try:
        ownership = _stat(file_name)
    except subprocess.CalledProcessError as e:
        print("Error reading file: {}".format(e))
        if not optional:
            assert False, "Specified file does not exist: {}".format(file_name)
    assert mode == ownership.mode, \
        "{} has an incorrect mode: {} should be {}".format(
            file_name, ownership.mode, mode)
    print("Validate mode of {}: PASS".format(file_name))


@cached
def _config_section(config, section):
    """Read the configuration file and return a section."""
    path = os.path.join(config.get('config_path'), config.get('config_file'))
    conf = _config_ini(path)
    return conf.get(section)


@audit(is_audit_type(AuditType.OpenStackSecurityGuide),
       it_has_config('files'))
def validate_file_ownership(config):
    """Verify that configuration files are owned by the correct user/group."""
    files = config.get('files', {})
    for file_name, options in files.items():
        for key in options.keys():
            if key not in ["owner", "group", "mode"]:
                raise RuntimeError(
                    "Invalid ownership configuration: {}".format(key))
        owner = options.get('owner', config.get('owner', 'root'))
        group = options.get('group', config.get('group', 'root'))
        optional = options.get('optional', config.get('optional', False))
        if '*' in file_name:
            for file in glob.glob(file_name):
                if file not in files.keys():
                    if os.path.isfile(file):
                        _validate_file_ownership(owner, group, file, optional)
        else:
            if os.path.isfile(file_name):
                _validate_file_ownership(owner, group, file_name, optional)


@audit(is_audit_type(AuditType.OpenStackSecurityGuide),
       it_has_config('files'))
def validate_file_permissions(config):
    """Verify that permissions on configuration files are secure enough."""
    files = config.get('files', {})
    for file_name, options in files.items():
        for key in options.keys():
            if key not in ["owner", "group", "mode"]:
                raise RuntimeError(
                    "Invalid ownership configuration: {}".format(key))
        mode = options.get('mode', config.get('permissions', '600'))
        optional = options.get('optional', config.get('optional', False))
        if '*' in file_name:
            for file in glob.glob(file_name):
                if file not in files.keys():
                    if os.path.isfile(file):
                        _validate_file_mode(mode, file, optional)
        else:
            if os.path.isfile(file_name):
                _validate_file_mode(mode, file_name, optional)


@audit(is_audit_type(AuditType.OpenStackSecurityGuide))
def validate_uses_keystone(audit_options):
    """Validate that the service uses Keystone for authentication."""
    section = _config_section(audit_options, 'api') or _config_section(audit_options, 'DEFAULT')
    assert section is not None, "Missing section 'api / DEFAULT'"
    assert section.get('auth_strategy') == "keystone", \
        "Application is not using Keystone"


@audit(is_audit_type(AuditType.OpenStackSecurityGuide))
def validate_uses_tls_for_keystone(audit_options):
    """Verify that TLS is used to communicate with Keystone."""
    section = _config_section(audit_options, 'keystone_authtoken')
    assert section is not None, "Missing section 'keystone_authtoken'"
    assert not section.get('insecure') and \
        "https://" in section.get("auth_uri"), \
        "TLS is not used for Keystone"


@audit(is_audit_type(AuditType.OpenStackSecurityGuide))
def validate_uses_tls_for_glance(audit_options):
    """Verify that TLS is used to communicate with Glance."""
    section = _config_section(audit_options, 'glance')
    assert section is not None, "Missing section 'glance'"
    assert not section.get('insecure') and \
        "https://" in section.get("api_servers"), \
        "TLS is not used for Glance"



# Copyright 2014-2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Remote Python Debugger (pdb wrapper)."""

import pdb
import socket
import sys

__author__ = "Bertrand Janin <b@janin.com>"
__version__ = "0.1.3"


class Rpdb(pdb.Pdb):

    def __init__(self, addr="127.0.0.1", port=4444):
        """Initialize the socket and initialize pdb."""

        # Backup stdin and stdout before replacing them by the socket handle
        self.old_stdout = sys.stdout
        self.old_stdin = sys.stdin

        # Open a 'reusable' socket to let the webapp reload on the same port
        self.skt = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.skt.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, True)
        self.skt.bind((addr, port))
        self.skt.listen(1)
        (clientsocket, address) = self.skt.accept()
        handle = clientsocket.makefile('rw')
        pdb.Pdb.__init__(self, completekey='tab', stdin=handle, stdout=handle)
        sys.stdout = sys.stdin = handle

    def shutdown(self):
        """Revert stdin and stdout, close the socket."""
        sys.stdout = self.old_stdout
        sys.stdin = self.old_stdin
        self.skt.close()
        self.set_continue()

    def do_continue(self, arg):
        """Stop all operation on ``continue``."""
        self.shutdown()
        return 1

    do_EOF = do_quit = do_exit = do_c = do_cont = do_continue



# Copyright 2018 Canonical Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import subprocess

from charmhelpers.core.hookenv import (
    ERROR,
    log,
    relation_get,
)
from charmhelpers.contrib.network.ip import (
    is_ipv6,
    ns_query,
)
from charmhelpers.contrib.openstack.utils import (
    get_hostname,
    get_host_ip,
    is_ip,
)

NOVA_SSH_DIR = '/etc/nova/compute_ssh/'


def ssh_directory_for_unit(application_name, user=None):
    """Return the directory used to store ssh assets for the application.

    :param application_name: Name of application eg nova-compute-something
    :type application_name: str
    :param user: The user that the ssh asserts are for.
    :type user: str
    :returns: Fully qualified directory path.
    :rtype: str
    """
    if user:
        application_name = "{}_{}".format(application_name, user)
    _dir = os.path.join(NOVA_SSH_DIR, application_name)
    for d in [NOVA_SSH_DIR, _dir]:
        if not os.path.isdir(d):
            os.mkdir(d)
    for f in ['authorized_keys', 'known_hosts']:
        f = os.path.join(_dir, f)
        if not os.path.isfile(f):
            open(f, 'w').close()
    return _dir


def known_hosts(application_name, user=None):
    """Return the known hosts file for the application.

    :param application_name: Name of application eg nova-compute-something
    :type application_name: str
    :param user: The user that the ssh asserts are for.
    :type user: str
    :returns: Fully qualified path to file.
    :rtype: str
    """
    return os.path.join(
        ssh_directory_for_unit(application_name, user),
        'known_hosts')


def authorized_keys(application_name, user=None):
    """Return the authorized keys file for the application.

    :param application_name: Name of application eg nova-compute-something
    :type application_name: str
    :param user: The user that the ssh asserts are for.
    :type user: str
    :returns: Fully qualified path to file.
    :rtype: str
    """
    return os.path.join(
        ssh_directory_for_unit(application_name, user),
        'authorized_keys')


def ssh_known_host_key(host, application_name, user=None):
    """Return the first entry in known_hosts for host.

    :param host: hostname to lookup in file.
    :type host: str
    :param application_name: Name of application eg nova-compute-something
    :type application_name: str
    :param user: The user that the ssh asserts are for.
    :type user: str
    :returns: Host key
    :rtype: str or None
    """
    cmd = [
        'ssh-keygen',
        '-f', known_hosts(application_name, user),
        '-H',
        '-F',
        host]
    try:
        # The first line of output is like '# Host xx found: line 1 type RSA',
        # which should be excluded.
        output = subprocess.check_output(cmd)
    except subprocess.CalledProcessError as e:
        # RC of 1 seems to be legitimate for most ssh-keygen -F calls.
        if e.returncode == 1:
            output = e.output
        else:
            raise
    output = output.strip()

    if output:
        # Bug #1500589 cmd has 0 rc on precise if entry not present
        lines = output.split('\n')
        if len(lines) >= 1:
            return lines[0]

    return None


def remove_known_host(host, application_name, user=None):
    """Remove the entry in known_hosts for host.

    :param host: hostname to lookup in file.
    :type host: str
    :param application_name: Name of application eg nova-compute-something
    :type application_name: str
    :param user: The user that the ssh asserts are for.
    :type user: str
    """
    log('Removing SSH known host entry for compute host at %s' % host)
    cmd = ['ssh-keygen', '-f', known_hosts(application_name, user), '-R', host]
    subprocess.check_call(cmd)


def is_same_key(key_1, key_2):
    """Extract the key from two host entries and compare them.

    :param key_1: Host key
    :type key_1: str
    :param key_2: Host key
    :type key_2: str
    """
    # The key format get will be like '|1|2rUumCavEXWVaVyB5uMl6m85pZo=|Cp'
    # 'EL6l7VTY37T/fg/ihhNb/GPgs= ssh-rsa AAAAB', we only need to compare
    # the part start with 'ssh-rsa' followed with '= ', because the hash
    # value in the beginning will change each time.
    k_1 = key_1.split('= ')[1]
    k_2 = key_2.split('= ')[1]
    return k_1 == k_2


def add_known_host(host, application_name, user=None):
    """Add the given host key to the known hosts file.

    :param host: host name
    :type host: str
    :param application_name: Name of application eg nova-compute-something
    :type application_name: str
    :param user: The user that the ssh asserts are for.
    :type user: str
    """
    cmd = ['ssh-keyscan', '-H', '-t', 'rsa', host]
    try:
        remote_key = subprocess.check_output(cmd).strip()
    except Exception as e:
        log('Could not obtain SSH host key from %s' % host, level=ERROR)
        raise e

    current_key = ssh_known_host_key(host, application_name, user)
    if current_key and remote_key:
        if is_same_key(remote_key, current_key):
            log('Known host key for compute host %s up to date.' % host)
            return
        else:
            remove_known_host(host, application_name, user)

    log('Adding SSH host key to known hosts for compute node at %s.' % host)
    with open(known_hosts(application_name, user), 'a') as out:
        out.write("{}\n".format(remote_key))


def ssh_authorized_key_exists(public_key, application_name, user=None):
    """Check if given key is in the authorized_key file.

    :param public_key: Public key.
    :type public_key: str
    :param application_name: Name of application eg nova-compute-something
    :type application_name: str
    :param user: The user that the ssh asserts are for.
    :type user: str
    :returns: Whether given key is in the authorized_key file.
    :rtype: boolean
    """
    with open(authorized_keys(application_name, user)) as keys:
        return ('%s' % public_key) in keys.read()


def add_authorized_key(public_key, application_name, user=None):
    """Add given key to the authorized_key file.

    :param public_key: Public key.
    :type public_key: str
    :param application_name: Name of application eg nova-compute-something
    :type application_name: str
    :param user: The user that the ssh asserts are for.
    :type user: str
    """
    with open(authorized_keys(application_name, user), 'a') as keys:
        keys.write("{}\n".format(public_key))


def ssh_compute_add_host_and_key(public_key, hostname, private_address,
                                 application_name, user=None):
    """Add a compute nodes ssh details to local cache.

    Collect various hostname variations and add the corresponding host keys to
    the local known hosts file. Finally, add the supplied public key to the
    authorized_key file.

    :param public_key: Public key.
    :type public_key: str
    :param hostname: Hostname to collect host keys from.
    :type hostname: str
    :param private_address:aCorresponding private address for hostname
    :type private_address: str
    :param application_name: Name of application eg nova-compute-something
    :type application_name: str
    :param user: The user that the ssh asserts are for.
    :type user: str
    """
    # If remote compute node hands us a hostname, ensure we have a
    # known hosts entry for its IP, hostname and FQDN.
    hosts = [private_address]

    if not is_ipv6(private_address):
        if hostname:
            hosts.append(hostname)

        if is_ip(private_address):
            hn = get_hostname(private_address)
            if hn:
                hosts.append(hn)
                short = hn.split('.')[0]
                if ns_query(short):
                    hosts.append(short)
        else:
            hosts.append(get_host_ip(private_address))
            short = private_address.split('.')[0]
            if ns_query(short):
                hosts.append(short)

    for host in list(set(hosts)):
        add_known_host(host, application_name, user)

    if not ssh_authorized_key_exists(public_key, application_name, user):
        log('Saving SSH authorized key for compute host at %s.' %
            private_address)
        add_authorized_key(public_key, application_name, user)


def ssh_compute_add(public_key, application_name, rid=None, unit=None,
                    user=None):
    """Add a compute nodes ssh details to local cache.

    Collect various hostname variations and add the corresponding host keys to
    the local known hosts file. Finally, add the supplied public key to the
    authorized_key file.

    :param public_key: Public key.
    :type public_key: str
    :param application_name: Name of application eg nova-compute-something
    :type application_name: str
    :param rid: Relation id of the relation between this charm and the app. If
                none is supplied it is assumed its the relation relating to
                the current hook context.
    :type rid: str
    :param unit: Unit to add ssh asserts for if none is supplied it is assumed
                 its the unit relating to the current hook context.
    :type unit: str
    :param user: The user that the ssh asserts are for.
    :type user: str
    """
    relation_data = relation_get(rid=rid, unit=unit)
    ssh_compute_add_host_and_key(
        public_key,
        relation_data.get('hostname'),
        relation_data.get('private-address'),
        application_name,
        user=user)


def ssh_known_hosts_lines(application_name, user=None):
    """Return contents of known_hosts file for given application.

    :param application_name: Name of application eg nova-compute-something
    :type application_name: str
    :param user: The user that the ssh asserts are for.
    :type user: str
    """
    known_hosts_list = []
    with open(known_hosts(application_name, user)) as hosts:
        for hosts_line in hosts:
            if hosts_line.rstrip():
                known_hosts_list.append(hosts_line.rstrip())
    return(known_hosts_list)


def ssh_authorized_keys_lines(application_name, user=None):
    """Return contents of authorized_keys file for given application.

    :param application_name: Name of application eg nova-compute-something
    :type application_name: str
    :param user: The user that the ssh asserts are for.
    :type user: str
    """
    authorized_keys_list = []

    with open(authorized_keys(application_name, user)) as keys:
        for authkey_line in keys:
            if authkey_line.rstrip():
                authorized_keys_list.append(authkey_line.rstrip())
    return(authorized_keys_list)


def ssh_compute_remove(public_key, application_name, user=None):
    """Remove given public key from authorized_keys file.

    :param public_key: Public key.
    :type public_key: str
    :param application_name: Name of application eg nova-compute-something
    :type application_name: str
    :param user: The user that the ssh asserts are for.
    :type user: str
    """
    if not (os.path.isfile(authorized_keys(application_name, user)) or
            os.path.isfile(known_hosts(application_name, user))):
        return

    keys = ssh_authorized_keys_lines(application_name, user=None)
    keys = [k.strip() for k in keys]

    if public_key not in keys:
        return

    [keys.remove(key) for key in keys if key == public_key]

    with open(authorized_keys(application_name, user), 'w') as _keys:
        keys = '\n'.join(keys)
        if not keys.endswith('\n'):
            keys += '\n'
        _keys.write(keys)


def get_ssh_settings(application_name, user=None):
    """Retrieve the known host entries and public keys for application

    Retrieve the known host entries and public keys for application for all
    units of the given application related to this application for the
    app + user combination.

    :param application_name: Name of application eg nova-compute-something
    :type application_name: str
    :param user: The user that the ssh asserts are for.
    :type user: str
    :returns: Public keys + host keys for all units for app + user combination.
    :rtype: dict
    """
    settings = {}
    keys = {}
    prefix = ''
    if user:
        prefix = '{}_'.format(user)

    for i, line in enumerate(ssh_known_hosts_lines(
            application_name=application_name, user=user)):
        settings['{}known_hosts_{}'.format(prefix, i)] = line
    if settings:
        settings['{}known_hosts_max_index'.format(prefix)] = len(
            settings.keys())

    for i, line in enumerate(ssh_authorized_keys_lines(
            application_name=application_name, user=user)):
        keys['{}authorized_keys_{}'.format(prefix, i)] = line
    if keys:
        keys['{}authorized_keys_max_index'.format(prefix)] = len(keys.keys())
    settings.update(keys)
    return settings


def get_all_user_ssh_settings(application_name):
    """Retrieve the known host entries and public keys for application

    Retrieve the known host entries and public keys for application for all
    units of the given application related to this application for root user
    and nova user.

    :param application_name: Name of application eg nova-compute-something
    :type application_name: str
    :returns: Public keys + host keys for all units for app + user combination.
    :rtype: dict
    """
    settings = get_ssh_settings(application_name)
    settings.update(get_ssh_settings(application_name, user='nova'))
    return settings



# Copyright 2014-2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from subprocess import check_output, CalledProcessError, STDOUT
from charmhelpers.fetch import (
    BaseFetchHandler,
    UnhandledSource,
    filter_installed_packages,
    install,
)

if filter_installed_packages(['git']) != []:
    install(['git'])
    if filter_installed_packages(['git']) != []:
        raise NotImplementedError('Unable to install git')


class GitUrlFetchHandler(BaseFetchHandler):
    """Handler for git branches via generic and github URLs."""

    def can_handle(self, source):
        url_parts = self.parse_url(source)
        # TODO (mattyw) no support for ssh git@ yet
        if url_parts.scheme not in ('http', 'https', 'git', ''):
            return False
        elif not url_parts.scheme:
            return os.path.exists(os.path.join(source, '.git'))
        else:
            return True

    def clone(self, source, dest, branch="master", depth=None):
        if not self.can_handle(source):
            raise UnhandledSource("Cannot handle {}".format(source))

        if os.path.exists(dest):
            cmd = ['git', '-C', dest, 'pull', source, branch]
        else:
            cmd = ['git', 'clone', source, dest, '--branch', branch]
            if depth:
                cmd.extend(['--depth', depth])
        check_output(cmd, stderr=STDOUT)

    def install(self, source, branch="master", dest=None, depth=None):
        url_parts = self.parse_url(source)
        branch_name = url_parts.path.strip("/").split("/")[-1]
        if dest:
            dest_dir = os.path.join(dest, branch_name)
        else:
            dest_dir = os.path.join(os.environ.get('CHARM_DIR'), "fetched",
                                    branch_name)
        try:
            self.clone(source, dest_dir, branch, depth)
        except CalledProcessError as e:
            raise UnhandledSource(e)
        except OSError as e:
            raise UnhandledSource(e.strerror)
        return dest_dir




from mock import call, patch, MagicMock, sentinel
from testtools import TestCase

from tests.helpers import patch_open, FakeRelation
from charmhelpers.contrib import unison


FAKE_RELATION_ENV = {
    'cluster:0': ['cluster/0', 'cluster/1']
}


TO_PATCH = [
    'log', 'check_call', 'check_output', 'relation_ids',
    'related_units', 'relation_get', 'relation_set',
    'hook_name', 'unit_private_ip',
]

FAKE_LOCAL_UNIT = 'test_host'
FAKE_RELATION = {
    'cluster:0': {
        'cluster/0': {
            'private-address': 'cluster0.local',
            'ssh_authorized_hosts': 'someotherhost:test_host'
        },
        'clsuter/1': {
            'private-address': 'cluster1.local',
            'ssh_authorized_hosts': 'someotherhost'
        },
        'clsuter/3': {
            'private-address': 'cluster2.local',
            'ssh_authorized_hosts': 'someotherthirdhost'
        },

    },

}


class UnisonHelperTests(TestCase):
    def setUp(self):
        super(UnisonHelperTests, self).setUp()
        for m in TO_PATCH:
            setattr(self, m, self._patch(m))
        self.fake_relation = FakeRelation(FAKE_RELATION)
        self.unit_private_ip.return_value = FAKE_LOCAL_UNIT
        self.relation_get.side_effect = self.fake_relation.get
        self.relation_ids.side_effect = self.fake_relation.relation_ids
        self.related_units.side_effect = self.fake_relation.related_units

    def _patch(self, method):
        _m = patch('charmhelpers.contrib.unison.' + method)
        mock = _m.start()
        self.addCleanup(_m.stop)
        return mock

    @patch('pwd.getpwnam')
    def test_get_homedir(self, pwnam):
        fake_user = MagicMock()
        fake_user.pw_dir = '/home/foo'
        pwnam.return_value = fake_user
        self.assertEquals(unison.get_homedir('foo'),
                          '/home/foo')

    @patch('pwd.getpwnam')
    def test_get_homedir_no_user(self, pwnam):
        e = KeyError
        pwnam.side_effect = e
        self.assertRaises(Exception, unison.get_homedir, user='foo')

    def _ensure_calls_in(self, calls):
        for _call in calls:
            self.assertIn(call(_call), self.check_call.call_args_list)

    @patch('os.path.isfile')
    def test_create_private_key_rsa(self, isfile):
        create_cmd = [
            'ssh-keygen', '-q', '-N', '', '-t', 'rsa', '-b', '2048',
            '-f', '/home/foo/.ssh/id_rsa']

        def _ensure_perms():
            cmds = [
                ['chown', 'foo', '/home/foo/.ssh/id_rsa'],
                ['chmod', '0600', '/home/foo/.ssh/id_rsa'],
            ]
            self._ensure_calls_in(cmds)

        isfile.return_value = False
        unison.create_private_key(
            user='foo', priv_key_path='/home/foo/.ssh/id_rsa')
        self.assertIn(call(create_cmd), self.check_call.call_args_list)
        _ensure_perms()
        self.check_call.call_args_list = []

        isfile.return_value = True
        unison.create_private_key(
            user='foo', priv_key_path='/home/foo/.ssh/id_rsa')
        self.assertNotIn(call(create_cmd), self.check_call.call_args_list)
        _ensure_perms()

    @patch('os.path.isfile')
    def test_create_private_key_ecdsa(self, isfile):
        create_cmd = [
            'ssh-keygen', '-q', '-N', '', '-t', 'ecdsa', '-b', '521',
            '-f', '/home/foo/.ssh/id_ecdsa']

        def _ensure_perms():
            cmds = [
                ['chown', 'foo', '/home/foo/.ssh/id_ecdsa'],
                ['chmod', '0600', '/home/foo/.ssh/id_ecdsa'],
            ]
            self._ensure_calls_in(cmds)

        isfile.return_value = False
        unison.create_private_key(
            user='foo',
            priv_key_path='/home/foo/.ssh/id_ecdsa',
            key_type='ecdsa')
        self.assertIn(call(create_cmd), self.check_call.call_args_list)
        _ensure_perms()
        self.check_call.call_args_list = []

        isfile.return_value = True
        unison.create_private_key(
            user='foo',
            priv_key_path='/home/foo/.ssh/id_ecdsa',
            key_type='ecdsa')
        self.assertNotIn(call(create_cmd), self.check_call.call_args_list)
        _ensure_perms()

    @patch('os.path.isfile')
    def test_create_public_key(self, isfile):
        create_cmd = ['ssh-keygen', '-y', '-f', '/home/foo/.ssh/id_rsa']
        isfile.return_value = True
        unison.create_public_key(
            user='foo', priv_key_path='/home/foo/.ssh/id_rsa',
            pub_key_path='/home/foo/.ssh/id_rsa.pub')
        self.assertNotIn(call(create_cmd), self.check_output.call_args_list)

        isfile.return_value = False
        with patch_open() as (_open, _file):
            self.check_output.return_value = b'fookey'
            unison.create_public_key(
                user='foo', priv_key_path='/home/foo/.ssh/id_rsa',
                pub_key_path='/home/foo/.ssh/id_rsa.pub')
            self.assertIn(call(create_cmd), self.check_output.call_args_list)
            _open.assert_called_with('/home/foo/.ssh/id_rsa.pub', 'wb')
            _file.write.assert_called_with(b'fookey')

    @patch('os.mkdir')
    @patch('os.path.isdir')
    @patch.object(unison, 'get_homedir')
    @patch.multiple(unison, create_private_key=MagicMock(),
                    create_public_key=MagicMock())
    def test_get_keypair(self, get_homedir, isdir, mkdir):
        get_homedir.return_value = '/home/foo'
        isdir.return_value = False
        with patch_open() as (_open, _file):
            _file.read.side_effect = [
                'foopriv', 'foopub'
            ]
            priv, pub = unison.get_keypair('adam')
            for f in ['/home/foo/.ssh/id_rsa',
                      '/home/foo/.ssh/id_rsa.pub']:
                self.assertIn(call(f, 'r'), _open.call_args_list)
        self.assertEquals(priv, 'foopriv')
        self.assertEquals(pub, 'foopub')

    @patch.object(unison, 'get_homedir')
    def test_write_auth_keys(self, get_homedir):
        get_homedir.return_value = '/home/foo'
        keys = [
            'ssh-rsa AAAB3Nz adam',
            'ssh-rsa ALKJFz adam@whereschuck.org',
        ]
        with patch_open() as (_open, _file):
            unison.write_authorized_keys('foo', keys)
            _open.assert_called_with('/home/foo/.ssh/authorized_keys', 'w')
            for k in keys:
                self.assertIn(call('%s\n' % k), _file.write.call_args_list)

    @patch.object(unison, 'get_homedir')
    def test_write_known_hosts(self, get_homedir):
        get_homedir.return_value = '/home/foo'
        keys = [
            '10.0.0.1 ssh-rsa KJDSJF=',
            '10.0.0.2 ssh-rsa KJDSJF=',
        ]
        self.check_output.side_effect = keys
        with patch_open() as (_open, _file):
            unison.write_known_hosts('foo', ['10.0.0.1', '10.0.0.2'])
            _open.assert_called_with('/home/foo/.ssh/known_hosts', 'w')
            for k in keys:
                self.assertIn(call('%s\n' % k), _file.write.call_args_list)

    @patch.object(unison, 'remove_password_expiry')
    @patch.object(unison, 'pwgen')
    @patch.object(unison, 'add_user_to_group')
    @patch.object(unison, 'adduser')
    def test_ensure_user(self, adduser, to_group, pwgen,
                         remove_password_expiry):
        pwgen.return_value = sentinel.password
        unison.ensure_user('foo', group='foobar')
        adduser.assert_called_with('foo', sentinel.password)
        to_group.assert_called_with('foo', 'foobar')
        remove_password_expiry.assert_called_with('foo')

    @patch.object(unison, '_run_as_user')
    def test_run_as_user(self, _run):
        with patch.object(unison, '_run_as_user') as _run:
            fake_preexec = MagicMock()
            _run.return_value = fake_preexec
            unison.run_as_user('foo', ['echo', 'foo'])
            self.check_output.assert_called_with(
                ['echo', 'foo'], preexec_fn=fake_preexec, cwd='/')

    @patch('pwd.getpwnam')
    def test_run_user_not_found(self, getpwnam):
        e = KeyError
        getpwnam.side_effect = e
        self.assertRaises(Exception, unison._run_as_user, 'nouser')

    @patch('os.setuid')
    @patch('os.setgid')
    @patch('os.environ', spec=dict)
    @patch('pwd.getpwnam')
    def test_run_as_user_preexec(self, pwnam, environ, setgid, setuid):
        fake_env = {'HOME': '/root'}
        environ.__getitem__ = MagicMock()
        environ.__setitem__ = MagicMock()
        environ.__setitem__.side_effect = fake_env.__setitem__
        environ.__getitem__.side_effect = fake_env.__getitem__

        fake_user = MagicMock()
        fake_user.pw_uid = 1010
        fake_user.pw_gid = 1011
        fake_user.pw_dir = '/home/foo'
        pwnam.return_value = fake_user
        inner = unison._run_as_user('foo')
        self.assertEquals(fake_env['HOME'], '/home/foo')
        inner()
        setgid.assert_called_with(1011)
        setuid.assert_called_with(1010)

    @patch('os.setuid')
    @patch('os.setgid')
    @patch('os.environ', spec=dict)
    @patch('pwd.getpwnam')
    def test_run_as_user_preexec_with_group(self, pwnam, environ, setgid, setuid):
        fake_env = {'HOME': '/root'}
        environ.__getitem__ = MagicMock()
        environ.__setitem__ = MagicMock()
        environ.__setitem__.side_effect = fake_env.__setitem__
        environ.__getitem__.side_effect = fake_env.__getitem__

        fake_user = MagicMock()
        fake_user.pw_uid = 1010
        fake_user.pw_gid = 1011
        fake_user.pw_dir = '/home/foo'
        fake_group_id = 2000
        pwnam.return_value = fake_user
        inner = unison._run_as_user('foo', gid=fake_group_id)
        self.assertEquals(fake_env['HOME'], '/home/foo')
        inner()
        setgid.assert_called_with(2000)
        setuid.assert_called_with(1010)

    @patch.object(unison, 'get_keypair')
    @patch.object(unison, 'ensure_user')
    def test_ssh_auth_peer_joined(self, ensure_user, get_keypair):
        get_keypair.return_value = ('privkey', 'pubkey')
        self.hook_name.return_value = 'cluster-relation-joined'
        unison.ssh_authorized_peers(peer_interface='cluster',
                                    user='foo', group='foo',
                                    ensure_local_user=True)
        self.relation_set.assert_called_with(ssh_pub_key='pubkey')
        self.assertFalse(self.relation_get.called)
        ensure_user.assert_called_with('foo', 'foo')
        get_keypair.assert_called_with('foo')

    @patch.object(unison, 'write_known_hosts')
    @patch.object(unison, 'write_authorized_keys')
    @patch.object(unison, 'get_keypair')
    @patch.object(unison, 'ensure_user')
    def test_ssh_auth_peer_changed(self, ensure_user, get_keypair,
                                   write_keys, write_hosts):
        get_keypair.return_value = ('privkey', 'pubkey')

        self.hook_name.return_value = 'cluster-relation-changed'

        self.relation_get.side_effect = [
            'key1',
            'host1',
            'key2',
            'host2',
            '', ''
        ]
        unison.ssh_authorized_peers(peer_interface='cluster',
                                    user='foo', group='foo',
                                    ensure_local_user=True)

        ensure_user.assert_called_with('foo', 'foo')
        get_keypair.assert_called_with('foo')
        write_keys.assert_called_with('foo', ['key1', 'key2'])
        write_hosts.assert_called_with('foo', ['host1', 'host2'])
        self.relation_set.assert_called_with(ssh_authorized_hosts='host1:host2')

    @patch.object(unison, 'write_known_hosts')
    @patch.object(unison, 'write_authorized_keys')
    @patch.object(unison, 'get_keypair')
    @patch.object(unison, 'ensure_user')
    def test_ssh_auth_peer_departed(self, ensure_user, get_keypair,
                                    write_keys, write_hosts):
        get_keypair.return_value = ('privkey', 'pubkey')

        self.hook_name.return_value = 'cluster-relation-departed'

        self.relation_get.side_effect = [
            'key1',
            'host1',
            'key2',
            'host2',
            '', ''
        ]
        unison.ssh_authorized_peers(peer_interface='cluster',
                                    user='foo', group='foo',
                                    ensure_local_user=True)

        ensure_user.assert_called_with('foo', 'foo')
        get_keypair.assert_called_with('foo')
        write_keys.assert_called_with('foo', ['key1', 'key2'])
        write_hosts.assert_called_with('foo', ['host1', 'host2'])
        self.relation_set.assert_called_with(ssh_authorized_hosts='host1:host2')

    def test_collect_authed_hosts(self):
        # only one of the hosts in fake environment has auth'd
        # the local peer
        hosts = unison.collect_authed_hosts('cluster')
        self.assertEquals(hosts, ['cluster0.local'])

    def test_collect_authed_hosts_none_authed(self):
        with patch.object(unison, 'relation_get') as relation_get:
            relation_get.return_value = ''
            hosts = unison.collect_authed_hosts('cluster')
            self.assertEquals(hosts, [])

    @patch.object(unison, 'run_as_user')
    def test_sync_path_to_host(self, run_as_user, verbose=True, gid=None):
        for path in ['/tmp/foo', '/tmp/foo/']:
            unison.sync_path_to_host(path=path, host='clusterhost1',
                                     user='foo', verbose=verbose, gid=gid)
            ex_cmd = ['unison', '-auto', '-batch=true',
                      '-confirmbigdel=false', '-fastcheck=true',
                      '-group=false', '-owner=false',
                      '-prefer=newer', '-times=true']
            if not verbose:
                ex_cmd.append('-silent')
            ex_cmd += ['/tmp/foo', 'ssh://foo@clusterhost1//tmp/foo']
            run_as_user.assert_called_with('foo', ex_cmd, gid)

    @patch.object(unison, 'run_as_user')
    def test_sync_path_to_host_error(self, run_as_user):
        for i, path in enumerate(['/tmp/foo', '/tmp/foo/']):
            run_as_user.side_effect = Exception
            if i == 0:
                unison.sync_path_to_host(path=path, host='clusterhost1',
                                         user='foo', verbose=True, gid=None)
            else:
                self.assertRaises(Exception, unison.sync_path_to_host,
                                  path=path, host='clusterhost1',
                                  user='foo', verbose=True, gid=None,
                                  fatal=True)

            ex_cmd = ['unison', '-auto', '-batch=true',
                      '-confirmbigdel=false', '-fastcheck=true',
                      '-group=false', '-owner=false',
                      '-prefer=newer', '-times=true',
                      '/tmp/foo', 'ssh://foo@clusterhost1//tmp/foo']
            run_as_user.assert_called_with('foo', ex_cmd, None)

    def test_sync_path_to_host_non_verbose(self):
        return self.test_sync_path_to_host(verbose=False)

    def test_sync_path_to_host_with_gid(self):
        return self.test_sync_path_to_host(gid=111)

    @patch.object(unison, 'sync_path_to_host')
    def test_sync_to_peer(self, sync_path_to_host):
        paths = ['/tmp/foo1', '/tmp/foo2']
        host = 'host1'
        unison.sync_to_peer(host, 'foouser', paths, True)
        calls = [call('/tmp/foo1', host, 'foouser', True, None, None, False),
                 call('/tmp/foo2', host, 'foouser', True, None, None, False)]
        sync_path_to_host.assert_has_calls(calls)

    @patch.object(unison, 'sync_path_to_host')
    def test_sync_to_peer_with_gid(self, sync_path_to_host):
        paths = ['/tmp/foo1', '/tmp/foo2']
        host = 'host1'
        unison.sync_to_peer(host, 'foouser', paths, True, gid=111)
        calls = [call('/tmp/foo1', host, 'foouser', True, None, 111, False),
                 call('/tmp/foo2', host, 'foouser', True, None, 111, False)]
        sync_path_to_host.assert_has_calls(calls)

    @patch.object(unison, 'collect_authed_hosts')
    @patch.object(unison, 'sync_to_peer')
    def test_sync_to_peers(self, sync_to_peer, collect_hosts):
        collect_hosts.return_value = ['host1', 'host2', 'host3']
        paths = ['/tmp/foo']
        unison.sync_to_peers(peer_interface='cluster', user='foouser',
                             paths=paths, verbose=True)
        calls = [call('host1', 'foouser', ['/tmp/foo'], True, None, None, False),
                 call('host2', 'foouser', ['/tmp/foo'], True, None, None, False),
                 call('host3', 'foouser', ['/tmp/foo'], True, None, None, False)]
        sync_to_peer.assert_has_calls(calls)

    @patch.object(unison, 'collect_authed_hosts')
    @patch.object(unison, 'sync_to_peer')
    def test_sync_to_peers_with_gid(self, sync_to_peer, collect_hosts):
        collect_hosts.return_value = ['host1', 'host2', 'host3']
        paths = ['/tmp/foo']
        unison.sync_to_peers(peer_interface='cluster', user='foouser',
                             paths=paths, verbose=True, gid=111)
        calls = [call('host1', 'foouser', ['/tmp/foo'], True, None, 111, False),
                 call('host2', 'foouser', ['/tmp/foo'], True, None, 111, False),
                 call('host3', 'foouser', ['/tmp/foo'], True, None, 111, False)]
        sync_to_peer.assert_has_calls(calls)

    @patch.object(unison, 'collect_authed_hosts')
    @patch.object(unison, 'sync_to_peer')
    def test_sync_to_peers_with_cmd(self, sync_to_peer, collect_hosts):
        collect_hosts.return_value = ['host1', 'host2', 'host3']
        paths = ['/tmp/foo']
        cmd = ['dummy_cmd']
        unison.sync_to_peers(peer_interface='cluster', user='foouser',
                             paths=paths, verbose=True, cmd=cmd, gid=111)
        calls = [call('host1', 'foouser', ['/tmp/foo'], True, cmd, 111, False),
                 call('host2', 'foouser', ['/tmp/foo'], True, cmd, 111, False),
                 call('host3', 'foouser', ['/tmp/foo'], True, cmd, 111, False)]
        sync_to_peer.assert_has_calls(calls)



# Copyright 2014-2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
import platform
import re
import six
import subprocess
import sys
import time

from charmhelpers.core.host import get_distrib_codename, get_system_env

from charmhelpers.core.hookenv import (
    log,
    DEBUG,
    WARNING,
    env_proxy_settings,
)
from charmhelpers.fetch import SourceConfigError, GPGKeyError
from charmhelpers.fetch import ubuntu_apt_pkg

PROPOSED_POCKET = (
    "# Proposed\n"
    "deb http://archive.ubuntu.com/ubuntu {}-proposed main universe "
    "multiverse restricted\n")
PROPOSED_PORTS_POCKET = (
    "# Proposed\n"
    "deb http://ports.ubuntu.com/ubuntu-ports {}-proposed main universe "
    "multiverse restricted\n")
# Only supports 64bit and ppc64 at the moment.
ARCH_TO_PROPOSED_POCKET = {
    'x86_64': PROPOSED_POCKET,
    'ppc64le': PROPOSED_PORTS_POCKET,
    'aarch64': PROPOSED_PORTS_POCKET,
    's390x': PROPOSED_PORTS_POCKET,
}
CLOUD_ARCHIVE_URL = "http://ubuntu-cloud.archive.canonical.com/ubuntu"
CLOUD_ARCHIVE_KEY_ID = '5EDB1B62EC4926EA'
CLOUD_ARCHIVE = """# Ubuntu Cloud Archive
deb http://ubuntu-cloud.archive.canonical.com/ubuntu {} main
"""
CLOUD_ARCHIVE_POCKETS = {
    # Folsom
    'folsom': 'precise-updates/folsom',
    'folsom/updates': 'precise-updates/folsom',
    'precise-folsom': 'precise-updates/folsom',
    'precise-folsom/updates': 'precise-updates/folsom',
    'precise-updates/folsom': 'precise-updates/folsom',
    'folsom/proposed': 'precise-proposed/folsom',
    'precise-folsom/proposed': 'precise-proposed/folsom',
    'precise-proposed/folsom': 'precise-proposed/folsom',
    # Grizzly
    'grizzly': 'precise-updates/grizzly',
    'grizzly/updates': 'precise-updates/grizzly',
    'precise-grizzly': 'precise-updates/grizzly',
    'precise-grizzly/updates': 'precise-updates/grizzly',
    'precise-updates/grizzly': 'precise-updates/grizzly',
    'grizzly/proposed': 'precise-proposed/grizzly',
    'precise-grizzly/proposed': 'precise-proposed/grizzly',
    'precise-proposed/grizzly': 'precise-proposed/grizzly',
    # Havana
    'havana': 'precise-updates/havana',
    'havana/updates': 'precise-updates/havana',
    'precise-havana': 'precise-updates/havana',
    'precise-havana/updates': 'precise-updates/havana',
    'precise-updates/havana': 'precise-updates/havana',
    'havana/proposed': 'precise-proposed/havana',
    'precise-havana/proposed': 'precise-proposed/havana',
    'precise-proposed/havana': 'precise-proposed/havana',
    # Icehouse
    'icehouse': 'precise-updates/icehouse',
    'icehouse/updates': 'precise-updates/icehouse',
    'precise-icehouse': 'precise-updates/icehouse',
    'precise-icehouse/updates': 'precise-updates/icehouse',
    'precise-updates/icehouse': 'precise-updates/icehouse',
    'icehouse/proposed': 'precise-proposed/icehouse',
    'precise-icehouse/proposed': 'precise-proposed/icehouse',
    'precise-proposed/icehouse': 'precise-proposed/icehouse',
    # Juno
    'juno': 'trusty-updates/juno',
    'juno/updates': 'trusty-updates/juno',
    'trusty-juno': 'trusty-updates/juno',
    'trusty-juno/updates': 'trusty-updates/juno',
    'trusty-updates/juno': 'trusty-updates/juno',
    'juno/proposed': 'trusty-proposed/juno',
    'trusty-juno/proposed': 'trusty-proposed/juno',
    'trusty-proposed/juno': 'trusty-proposed/juno',
    # Kilo
    'kilo': 'trusty-updates/kilo',
    'kilo/updates': 'trusty-updates/kilo',
    'trusty-kilo': 'trusty-updates/kilo',
    'trusty-kilo/updates': 'trusty-updates/kilo',
    'trusty-updates/kilo': 'trusty-updates/kilo',
    'kilo/proposed': 'trusty-proposed/kilo',
    'trusty-kilo/proposed': 'trusty-proposed/kilo',
    'trusty-proposed/kilo': 'trusty-proposed/kilo',
    # Liberty
    'liberty': 'trusty-updates/liberty',
    'liberty/updates': 'trusty-updates/liberty',
    'trusty-liberty': 'trusty-updates/liberty',
    'trusty-liberty/updates': 'trusty-updates/liberty',
    'trusty-updates/liberty': 'trusty-updates/liberty',
    'liberty/proposed': 'trusty-proposed/liberty',
    'trusty-liberty/proposed': 'trusty-proposed/liberty',
    'trusty-proposed/liberty': 'trusty-proposed/liberty',
    # Mitaka
    'mitaka': 'trusty-updates/mitaka',
    'mitaka/updates': 'trusty-updates/mitaka',
    'trusty-mitaka': 'trusty-updates/mitaka',
    'trusty-mitaka/updates': 'trusty-updates/mitaka',
    'trusty-updates/mitaka': 'trusty-updates/mitaka',
    'mitaka/proposed': 'trusty-proposed/mitaka',
    'trusty-mitaka/proposed': 'trusty-proposed/mitaka',
    'trusty-proposed/mitaka': 'trusty-proposed/mitaka',
    # Newton
    'newton': 'xenial-updates/newton',
    'newton/updates': 'xenial-updates/newton',
    'xenial-newton': 'xenial-updates/newton',
    'xenial-newton/updates': 'xenial-updates/newton',
    'xenial-updates/newton': 'xenial-updates/newton',
    'newton/proposed': 'xenial-proposed/newton',
    'xenial-newton/proposed': 'xenial-proposed/newton',
    'xenial-proposed/newton': 'xenial-proposed/newton',
    # Ocata
    'ocata': 'xenial-updates/ocata',
    'ocata/updates': 'xenial-updates/ocata',
    'xenial-ocata': 'xenial-updates/ocata',
    'xenial-ocata/updates': 'xenial-updates/ocata',
    'xenial-updates/ocata': 'xenial-updates/ocata',
    'ocata/proposed': 'xenial-proposed/ocata',
    'xenial-ocata/proposed': 'xenial-proposed/ocata',
    'xenial-proposed/ocata': 'xenial-proposed/ocata',
    # Pike
    'pike': 'xenial-updates/pike',
    'xenial-pike': 'xenial-updates/pike',
    'xenial-pike/updates': 'xenial-updates/pike',
    'xenial-updates/pike': 'xenial-updates/pike',
    'pike/proposed': 'xenial-proposed/pike',
    'xenial-pike/proposed': 'xenial-proposed/pike',
    'xenial-proposed/pike': 'xenial-proposed/pike',
    # Queens
    'queens': 'xenial-updates/queens',
    'xenial-queens': 'xenial-updates/queens',
    'xenial-queens/updates': 'xenial-updates/queens',
    'xenial-updates/queens': 'xenial-updates/queens',
    'queens/proposed': 'xenial-proposed/queens',
    'xenial-queens/proposed': 'xenial-proposed/queens',
    'xenial-proposed/queens': 'xenial-proposed/queens',
    # Rocky
    'rocky': 'bionic-updates/rocky',
    'bionic-rocky': 'bionic-updates/rocky',
    'bionic-rocky/updates': 'bionic-updates/rocky',
    'bionic-updates/rocky': 'bionic-updates/rocky',
    'rocky/proposed': 'bionic-proposed/rocky',
    'bionic-rocky/proposed': 'bionic-proposed/rocky',
    'bionic-proposed/rocky': 'bionic-proposed/rocky',
    # Stein
    'stein': 'bionic-updates/stein',
    'bionic-stein': 'bionic-updates/stein',
    'bionic-stein/updates': 'bionic-updates/stein',
    'bionic-updates/stein': 'bionic-updates/stein',
    'stein/proposed': 'bionic-proposed/stein',
    'bionic-stein/proposed': 'bionic-proposed/stein',
    'bionic-proposed/stein': 'bionic-proposed/stein',
    # Train
    'train': 'bionic-updates/train',
    'bionic-train': 'bionic-updates/train',
    'bionic-train/updates': 'bionic-updates/train',
    'bionic-updates/train': 'bionic-updates/train',
    'train/proposed': 'bionic-proposed/train',
    'bionic-train/proposed': 'bionic-proposed/train',
    'bionic-proposed/train': 'bionic-proposed/train',
    # Ussuri
    'ussuri': 'bionic-updates/ussuri',
    'bionic-ussuri': 'bionic-updates/ussuri',
    'bionic-ussuri/updates': 'bionic-updates/ussuri',
    'bionic-updates/ussuri': 'bionic-updates/ussuri',
    'ussuri/proposed': 'bionic-proposed/ussuri',
    'bionic-ussuri/proposed': 'bionic-proposed/ussuri',
    'bionic-proposed/ussuri': 'bionic-proposed/ussuri',
}


APT_NO_LOCK = 100  # The return code for "couldn't acquire lock" in APT.
CMD_RETRY_DELAY = 10  # Wait 10 seconds between command retries.
CMD_RETRY_COUNT = 3  # Retry a failing fatal command X times.


def filter_installed_packages(packages):
    """Return a list of packages that require installation."""
    cache = apt_cache()
    _pkgs = []
    for package in packages:
        try:
            p = cache[package]
            p.current_ver or _pkgs.append(package)
        except KeyError:
            log('Package {} has no installation candidate.'.format(package),
                level='WARNING')
            _pkgs.append(package)
    return _pkgs


def filter_missing_packages(packages):
    """Return a list of packages that are installed.

    :param packages: list of packages to evaluate.
    :returns list: Packages that are installed.
    """
    return list(
        set(packages) -
        set(filter_installed_packages(packages))
    )


def apt_cache(*_, **__):
    """Shim returning an object simulating the apt_pkg Cache.

    :param _: Accept arguments for compability, not used.
    :type _: any
    :param __: Accept keyword arguments for compability, not used.
    :type __: any
    :returns:Object used to interrogate the system apt and dpkg databases.
    :rtype:ubuntu_apt_pkg.Cache
    """
    if 'apt_pkg' in sys.modules:
        # NOTE(fnordahl): When our consumer use the upstream ``apt_pkg`` module
        # in conjunction with the apt_cache helper function, they may expect us
        # to call ``apt_pkg.init()`` for them.
        #
        # Detect this situation, log a warning and make the call to
        # ``apt_pkg.init()`` to avoid the consumer Python interpreter from
        # crashing with a segmentation fault.
        log('Support for use of upstream ``apt_pkg`` module in conjunction'
            'with charm-helpers is deprecated since 2019-06-25', level=WARNING)
        sys.modules['apt_pkg'].init()
    return ubuntu_apt_pkg.Cache()


def apt_install(packages, options=None, fatal=False):
    """Install one or more packages.

    :param packages: Package(s) to install
    :type packages: Option[str, List[str]]
    :param options: Options to pass on to apt-get
    :type options: Option[None, List[str]]
    :param fatal: Whether the command's output should be checked and
                  retried.
    :type fatal: bool
    :raises: subprocess.CalledProcessError
    """
    if options is None:
        options = ['--option=Dpkg::Options::=--force-confold']

    cmd = ['apt-get', '--assume-yes']
    cmd.extend(options)
    cmd.append('install')
    if isinstance(packages, six.string_types):
        cmd.append(packages)
    else:
        cmd.extend(packages)
    log("Installing {} with options: {}".format(packages,
                                                options))
    _run_apt_command(cmd, fatal)


def apt_upgrade(options=None, fatal=False, dist=False):
    """Upgrade all packages.

    :param options: Options to pass on to apt-get
    :type options: Option[None, List[str]]
    :param fatal: Whether the command's output should be checked and
                  retried.
    :type fatal: bool
    :param dist: Whether ``dist-upgrade`` should be used over ``upgrade``
    :type dist: bool
    :raises: subprocess.CalledProcessError
    """
    if options is None:
        options = ['--option=Dpkg::Options::=--force-confold']

    cmd = ['apt-get', '--assume-yes']
    cmd.extend(options)
    if dist:
        cmd.append('dist-upgrade')
    else:
        cmd.append('upgrade')
    log("Upgrading with options: {}".format(options))
    _run_apt_command(cmd, fatal)


def apt_update(fatal=False):
    """Update local apt cache."""
    cmd = ['apt-get', 'update']
    _run_apt_command(cmd, fatal)


def apt_purge(packages, fatal=False):
    """Purge one or more packages.

    :param packages: Package(s) to install
    :type packages: Option[str, List[str]]
    :param fatal: Whether the command's output should be checked and
                  retried.
    :type fatal: bool
    :raises: subprocess.CalledProcessError
    """
    cmd = ['apt-get', '--assume-yes', 'purge']
    if isinstance(packages, six.string_types):
        cmd.append(packages)
    else:
        cmd.extend(packages)
    log("Purging {}".format(packages))
    _run_apt_command(cmd, fatal)


def apt_autoremove(purge=True, fatal=False):
    """Purge one or more packages.
    :param purge: Whether the ``--purge`` option should be passed on or not.
    :type purge: bool
    :param fatal: Whether the command's output should be checked and
                  retried.
    :type fatal: bool
    :raises: subprocess.CalledProcessError
    """
    cmd = ['apt-get', '--assume-yes', 'autoremove']
    if purge:
        cmd.append('--purge')
    _run_apt_command(cmd, fatal)


def apt_mark(packages, mark, fatal=False):
    """Flag one or more packages using apt-mark."""
    log("Marking {} as {}".format(packages, mark))
    cmd = ['apt-mark', mark]
    if isinstance(packages, six.string_types):
        cmd.append(packages)
    else:
        cmd.extend(packages)

    if fatal:
        subprocess.check_call(cmd, universal_newlines=True)
    else:
        subprocess.call(cmd, universal_newlines=True)


def apt_hold(packages, fatal=False):
    return apt_mark(packages, 'hold', fatal=fatal)


def apt_unhold(packages, fatal=False):
    return apt_mark(packages, 'unhold', fatal=fatal)


def import_key(key):
    """Import an ASCII Armor key.

    A Radix64 format keyid is also supported for backwards
    compatibility. In this case Ubuntu keyserver will be
    queried for a key via HTTPS by its keyid. This method
    is less preferrable because https proxy servers may
    require traffic decryption which is equivalent to a
    man-in-the-middle attack (a proxy server impersonates
    keyserver TLS certificates and has to be explicitly
    trusted by the system).

    :param key: A GPG key in ASCII armor format,
                  including BEGIN and END markers or a keyid.
    :type key: (bytes, str)
    :raises: GPGKeyError if the key could not be imported
    """
    key = key.strip()
    if '-' in key or '\n' in key:
        # Send everything not obviously a keyid to GPG to import, as
        # we trust its validation better than our own. eg. handling
        # comments before the key.
        log("PGP key found (looks like ASCII Armor format)", level=DEBUG)
        if ('-----BEGIN PGP PUBLIC KEY BLOCK-----' in key and
                '-----END PGP PUBLIC KEY BLOCK-----' in key):
            log("Writing provided PGP key in the binary format", level=DEBUG)
            if six.PY3:
                key_bytes = key.encode('utf-8')
            else:
                key_bytes = key
            key_name = _get_keyid_by_gpg_key(key_bytes)
            key_gpg = _dearmor_gpg_key(key_bytes)
            _write_apt_gpg_keyfile(key_name=key_name, key_material=key_gpg)
        else:
            raise GPGKeyError("ASCII armor markers missing from GPG key")
    else:
        log("PGP key found (looks like Radix64 format)", level=WARNING)
        log("SECURELY importing PGP key from keyserver; "
            "full key not provided.", level=WARNING)
        # as of bionic add-apt-repository uses curl with an HTTPS keyserver URL
        # to retrieve GPG keys. `apt-key adv` command is deprecated as is
        # apt-key in general as noted in its manpage. See lp:1433761 for more
        # history. Instead, /etc/apt/trusted.gpg.d is used directly to drop
        # gpg
        key_asc = _get_key_by_keyid(key)
        # write the key in GPG format so that apt-key list shows it
        key_gpg = _dearmor_gpg_key(key_asc)
        _write_apt_gpg_keyfile(key_name=key, key_material=key_gpg)


def _get_keyid_by_gpg_key(key_material):
    """Get a GPG key fingerprint by GPG key material.
    Gets a GPG key fingerprint (40-digit, 160-bit) by the ASCII armor-encoded
    or binary GPG key material. Can be used, for example, to generate file
    names for keys passed via charm options.

    :param key_material: ASCII armor-encoded or binary GPG key material
    :type key_material: bytes
    :raises: GPGKeyError if invalid key material has been provided
    :returns: A GPG key fingerprint
    :rtype: str
    """
    # Use the same gpg command for both Xenial and Bionic
    cmd = 'gpg --with-colons --with-fingerprint'
    ps = subprocess.Popen(cmd.split(),
                          stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE,
                          stdin=subprocess.PIPE)
    out, err = ps.communicate(input=key_material)
    if six.PY3:
        out = out.decode('utf-8')
        err = err.decode('utf-8')
    if 'gpg: no valid OpenPGP data found.' in err:
        raise GPGKeyError('Invalid GPG key material provided')
    # from gnupg2 docs: fpr :: Fingerprint (fingerprint is in field 10)
    return re.search(r"^fpr:{9}([0-9A-F]{40}):$", out, re.MULTILINE).group(1)


def _get_key_by_keyid(keyid):
    """Get a key via HTTPS from the Ubuntu keyserver.
    Different key ID formats are supported by SKS keyservers (the longer ones
    are more secure, see "dead beef attack" and https://evil32.com/). Since
    HTTPS is used, if SSLBump-like HTTPS proxies are in place, they will
    impersonate keyserver.ubuntu.com and generate a certificate with
    keyserver.ubuntu.com in the CN field or in SubjAltName fields of a
    certificate. If such proxy behavior is expected it is necessary to add the
    CA certificate chain containing the intermediate CA of the SSLBump proxy to
    every machine that this code runs on via ca-certs cloud-init directive (via
    cloudinit-userdata model-config) or via other means (such as through a
    custom charm option). Also note that DNS resolution for the hostname in a
    URL is done at a proxy server - not at the client side.

    8-digit (32 bit) key ID
    https://keyserver.ubuntu.com/pks/lookup?search=0x4652B4E6
    16-digit (64 bit) key ID
    https://keyserver.ubuntu.com/pks/lookup?search=0x6E85A86E4652B4E6
    40-digit key ID:
    https://keyserver.ubuntu.com/pks/lookup?search=0x35F77D63B5CEC106C577ED856E85A86E4652B4E6

    :param keyid: An 8, 16 or 40 hex digit keyid to find a key for
    :type keyid: (bytes, str)
    :returns: A key material for the specified GPG key id
    :rtype: (str, bytes)
    :raises: subprocess.CalledProcessError
    """
    # options=mr - machine-readable output (disables html wrappers)
    keyserver_url = ('https://keyserver.ubuntu.com'
                     '/pks/lookup?op=get&options=mr&exact=on&search=0x{}')
    curl_cmd = ['curl', keyserver_url.format(keyid)]
    # use proxy server settings in order to retrieve the key
    return subprocess.check_output(curl_cmd,
                                   env=env_proxy_settings(['https']))


def _dearmor_gpg_key(key_asc):
    """Converts a GPG key in the ASCII armor format to the binary format.

    :param key_asc: A GPG key in ASCII armor format.
    :type key_asc: (str, bytes)
    :returns: A GPG key in binary format
    :rtype: (str, bytes)
    :raises: GPGKeyError
    """
    ps = subprocess.Popen(['gpg', '--dearmor'],
                          stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE,
                          stdin=subprocess.PIPE)
    out, err = ps.communicate(input=key_asc)
    # no need to decode output as it is binary (invalid utf-8), only error
    if six.PY3:
        err = err.decode('utf-8')
    if 'gpg: no valid OpenPGP data found.' in err:
        raise GPGKeyError('Invalid GPG key material. Check your network setup'
                          ' (MTU, routing, DNS) and/or proxy server settings'
                          ' as well as destination keyserver status.')
    else:
        return out


def _write_apt_gpg_keyfile(key_name, key_material):
    """Writes GPG key material into a file at a provided path.

    :param key_name: A key name to use for a key file (could be a fingerprint)
    :type key_name: str
    :param key_material: A GPG key material (binary)
    :type key_material: (str, bytes)
    """
    with open('/etc/apt/trusted.gpg.d/{}.gpg'.format(key_name),
              'wb') as keyf:
        keyf.write(key_material)


def add_source(source, key=None, fail_invalid=False):
    """Add a package source to this system.

    @param source: a URL or sources.list entry, as supported by
    add-apt-repository(1). Examples::

        ppa:charmers/example
        deb https://stub:key@private.example.com/ubuntu trusty main

    In addition:
        'proposed:' may be used to enable the standard 'proposed'
        pocket for the release.
        'cloud:' may be used to activate official cloud archive pockets,
        such as 'cloud:icehouse'
        'distro' may be used as a noop

    Full list of source specifications supported by the function are:

    'distro': A NOP; i.e. it has no effect.
    'proposed': the proposed deb spec [2] is wrtten to
      /etc/apt/sources.list/proposed
    'distro-proposed': adds <version>-proposed to the debs [2]
    'ppa:<ppa-name>': add-apt-repository --yes <ppa_name>
    'deb <deb-spec>': add-apt-repository --yes deb <deb-spec>
    'http://....': add-apt-repository --yes http://...
    'cloud-archive:<spec>': add-apt-repository -yes cloud-archive:<spec>
    'cloud:<release>[-staging]': specify a Cloud Archive pocket <release> with
      optional staging version.  If staging is used then the staging PPA [2]
      with be used.  If staging is NOT used then the cloud archive [3] will be
      added, and the 'ubuntu-cloud-keyring' package will be added for the
      current distro.

    Otherwise the source is not recognised and this is logged to the juju log.
    However, no error is raised, unless sys_error_on_exit is True.

    [1] deb http://ubuntu-cloud.archive.canonical.com/ubuntu {} main
        where {} is replaced with the derived pocket name.
    [2] deb http://archive.ubuntu.com/ubuntu {}-proposed \
        main universe multiverse restricted
        where {} is replaced with the lsb_release codename (e.g. xenial)
    [3] deb http://ubuntu-cloud.archive.canonical.com/ubuntu <pocket>
        to /etc/apt/sources.list.d/cloud-archive-list

    @param key: A key to be added to the system's APT keyring and used
    to verify the signatures on packages. Ideally, this should be an
    ASCII format GPG public key including the block headers. A GPG key
    id may also be used, but be aware that only insecure protocols are
    available to retrieve the actual public key from a public keyserver
    placing your Juju environment at risk. ppa and cloud archive keys
    are securely added automtically, so sould not be provided.

    @param fail_invalid: (boolean) if True, then the function raises a
    SourceConfigError is there is no matching installation source.

    @raises SourceConfigError() if for cloud:<pocket>, the <pocket> is not a
    valid pocket in CLOUD_ARCHIVE_POCKETS
    """
    _mapping = OrderedDict([
        (r"^distro$", lambda: None),  # This is a NOP
        (r"^(?:proposed|distro-proposed)$", _add_proposed),
        (r"^cloud-archive:(.*)$", _add_apt_repository),
        (r"^((?:deb |http:|https:|ppa:).*)$", _add_apt_repository),
        (r"^cloud:(.*)-(.*)\/staging$", _add_cloud_staging),
        (r"^cloud:(.*)-(.*)$", _add_cloud_distro_check),
        (r"^cloud:(.*)$", _add_cloud_pocket),
        (r"^snap:.*-(.*)-(.*)$", _add_cloud_distro_check),
    ])
    if source is None:
        source = ''
    for r, fn in six.iteritems(_mapping):
        m = re.match(r, source)
        if m:
            if key:
                # Import key before adding the source which depends on it,
                # as refreshing packages could fail otherwise.
                try:
                    import_key(key)
                except GPGKeyError as e:
                    raise SourceConfigError(str(e))
            # call the associated function with the captured groups
            # raises SourceConfigError on error.
            fn(*m.groups())
            break
    else:
        # nothing matched.  log an error and maybe sys.exit
        err = "Unknown source: {!r}".format(source)
        log(err)
        if fail_invalid:
            raise SourceConfigError(err)


def _add_proposed():
    """Add the PROPOSED_POCKET as /etc/apt/source.list.d/proposed.list

    Uses get_distrib_codename to determine the correct stanza for
    the deb line.

    For intel architecutres PROPOSED_POCKET is used for the release, but for
    other architectures PROPOSED_PORTS_POCKET is used for the release.
    """
    release = get_distrib_codename()
    arch = platform.machine()
    if arch not in six.iterkeys(ARCH_TO_PROPOSED_POCKET):
        raise SourceConfigError("Arch {} not supported for (distro-)proposed"
                                .format(arch))
    with open('/etc/apt/sources.list.d/proposed.list', 'w') as apt:
        apt.write(ARCH_TO_PROPOSED_POCKET[arch].format(release))


def _add_apt_repository(spec):
    """Add the spec using add_apt_repository

    :param spec: the parameter to pass to add_apt_repository
    :type spec: str
    """
    if '{series}' in spec:
        series = get_distrib_codename()
        spec = spec.replace('{series}', series)
    # software-properties package for bionic properly reacts to proxy settings
    # passed as environment variables (See lp:1433761). This is not the case
    # LTS and non-LTS releases below bionic.
    _run_with_retries(['add-apt-repository', '--yes', spec],
                      cmd_env=env_proxy_settings(['https']))


def _add_cloud_pocket(pocket):
    """Add a cloud pocket as /etc/apt/sources.d/cloud-archive.list

    Note that this overwrites the existing file if there is one.

    This function also converts the simple pocket in to the actual pocket using
    the CLOUD_ARCHIVE_POCKETS mapping.

    :param pocket: string representing the pocket to add a deb spec for.
    :raises: SourceConfigError if the cloud pocket doesn't exist or the
        requested release doesn't match the current distro version.
    """
    apt_install(filter_installed_packages(['ubuntu-cloud-keyring']),
                fatal=True)
    if pocket not in CLOUD_ARCHIVE_POCKETS:
        raise SourceConfigError(
            'Unsupported cloud: source option %s' %
            pocket)
    actual_pocket = CLOUD_ARCHIVE_POCKETS[pocket]
    with open('/etc/apt/sources.list.d/cloud-archive.list', 'w') as apt:
        apt.write(CLOUD_ARCHIVE.format(actual_pocket))


def _add_cloud_staging(cloud_archive_release, openstack_release):
    """Add the cloud staging repository which is in
    ppa:ubuntu-cloud-archive/<openstack_release>-staging

    This function checks that the cloud_archive_release matches the current
    codename for the distro that charm is being installed on.

    :param cloud_archive_release: string, codename for the release.
    :param openstack_release: String, codename for the openstack release.
    :raises: SourceConfigError if the cloud_archive_release doesn't match the
        current version of the os.
    """
    _verify_is_ubuntu_rel(cloud_archive_release, openstack_release)
    ppa = 'ppa:ubuntu-cloud-archive/{}-staging'.format(openstack_release)
    cmd = 'add-apt-repository -y {}'.format(ppa)
    _run_with_retries(cmd.split(' '))


def _add_cloud_distro_check(cloud_archive_release, openstack_release):
    """Add the cloud pocket, but also check the cloud_archive_release against
    the current distro, and use the openstack_release as the full lookup.

    This just calls _add_cloud_pocket() with the openstack_release as pocket
    to get the correct cloud-archive.list for dpkg to work with.

    :param cloud_archive_release:String, codename for the distro release.
    :param openstack_release: String, spec for the release to look up in the
        CLOUD_ARCHIVE_POCKETS
    :raises: SourceConfigError if this is the wrong distro, or the pocket spec
        doesn't exist.
    """
    _verify_is_ubuntu_rel(cloud_archive_release, openstack_release)
    _add_cloud_pocket("{}-{}".format(cloud_archive_release, openstack_release))


def _verify_is_ubuntu_rel(release, os_release):
    """Verify that the release is in the same as the current ubuntu release.

    :param release: String, lowercase for the release.
    :param os_release: String, the os_release being asked for
    :raises: SourceConfigError if the release is not the same as the ubuntu
        release.
    """
    ubuntu_rel = get_distrib_codename()
    if release != ubuntu_rel:
        raise SourceConfigError(
            'Invalid Cloud Archive release specified: {}-{} on this Ubuntu'
            'version ({})'.format(release, os_release, ubuntu_rel))


def _run_with_retries(cmd, max_retries=CMD_RETRY_COUNT, retry_exitcodes=(1,),
                      retry_message="", cmd_env=None):
    """Run a command and retry until success or max_retries is reached.

    :param cmd: The apt command to run.
    :type cmd: str
    :param max_retries: The number of retries to attempt on a fatal
                        command. Defaults to CMD_RETRY_COUNT.
    :type max_retries: int
    :param retry_exitcodes: Optional additional exit codes to retry.
                            Defaults to retry on exit code 1.
    :type retry_exitcodes: tuple
    :param retry_message: Optional log prefix emitted during retries.
    :type retry_message: str
    :param: cmd_env: Environment variables to add to the command run.
    :type cmd_env: Option[None, Dict[str, str]]
    """
    env = get_apt_dpkg_env()
    if cmd_env:
        env.update(cmd_env)

    if not retry_message:
        retry_message = "Failed executing '{}'".format(" ".join(cmd))
    retry_message += ". Will retry in {} seconds".format(CMD_RETRY_DELAY)

    retry_count = 0
    result = None

    retry_results = (None,) + retry_exitcodes
    while result in retry_results:
        try:
            result = subprocess.check_call(cmd, env=env)
        except subprocess.CalledProcessError as e:
            retry_count = retry_count + 1
            if retry_count > max_retries:
                raise
            result = e.returncode
            log(retry_message)
            time.sleep(CMD_RETRY_DELAY)


def _run_apt_command(cmd, fatal=False):
    """Run an apt command with optional retries.

    :param cmd: The apt command to run.
    :type cmd: str
    :param fatal: Whether the command's output should be checked and
                  retried.
    :type fatal: bool
    """
    if fatal:
        _run_with_retries(
            cmd, retry_exitcodes=(1, APT_NO_LOCK,),
            retry_message="Couldn't acquire DPKG lock")
    else:
        subprocess.call(cmd, env=get_apt_dpkg_env())


def get_upstream_version(package):
    """Determine upstream version based on installed package

    @returns None (if not installed) or the upstream version
    """
    cache = apt_cache()
    try:
        pkg = cache[package]
    except Exception:
        # the package is unknown to the current apt cache.
        return None

    if not pkg.current_ver:
        # package is known, but no version is currently installed.
        return None

    return ubuntu_apt_pkg.upstream_version(pkg.current_ver.ver_str)


def get_apt_dpkg_env():
    """Get environment suitable for execution of APT and DPKG tools.

    We keep this in a helper function instead of in a global constant to
    avoid execution on import of the library.
    :returns: Environment suitable for execution of APT and DPKG tools.
    :rtype: Dict[str, str]
    """
    # The fallback is used in the event of ``/etc/environment`` not containing
    # avalid PATH variable.
    return {'DEBIAN_FRONTEND': 'noninteractive',
            'PATH': get_system_env('PATH', '/usr/sbin:/usr/bin:/sbin:/bin')}



# Copyright 2016 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from unittest import TestCase

from mock import patch

from charmhelpers.contrib.hardening.host.checks import profile


class ProfileTestCase(TestCase):

    def setUp(self):
        super(ProfileTestCase, self).setUp()

        os.environ['JUJU_CHARM_DIR'] = '/tmp'
        self.addCleanup(lambda: os.environ.pop('JUJU_CHARM_DIR'))

    @patch.object(profile.utils, 'get_settings', lambda x:
                  {'security': {'kernel_enable_core_dump': False, 'ssh_tmout': False}})
    def test_core_dump_disabled(self):
        audits = profile.get_audits()
        self.assertEqual(1, len(audits))
        self.assertTrue(isinstance(audits[0], profile.TemplatedFile))

    @patch.object(profile.utils, 'get_settings', lambda x: {
        'security': {'kernel_enable_core_dump': True, 'ssh_tmout': False}
    })
    def test_core_dump_enabled(self):
        audits = profile.get_audits()
        self.assertEqual(0, len(audits))

    @patch.object(profile.utils, 'get_settings', lambda x:
                  {'security': {'kernel_enable_core_dump': True, 'ssh_tmout': False}})
    def test_ssh_tmout_disabled(self):
        audits = profile.get_audits()
        self.assertEqual(0, len(audits))

    @patch.object(profile.utils, 'get_settings', lambda x: {
        'security': {'kernel_enable_core_dump': True, 'ssh_tmout': 300}
    })
    def test_ssh_tmout_enabled(self):
        audits = profile.get_audits()
        self.assertEqual(1, len(audits))
        self.assertTrue(isinstance(audits[0], profile.TemplatedFile))

    @patch.object(profile.utils, 'log', lambda *args, **kwargs: None)
    def test_ProfileContext(self):
        ctxt = profile.ProfileContext()
        self.assertEqual(ctxt(), {
            'ssh_tmout': 300
        })



# Copyright 2015 Canonical Ltd.
#
# Authors:
#  Adam Collard <adam.collard@canonical.com>

from contextlib import contextmanager
from mock import patch
import sys
import unittest

import six

from charmhelpers.contrib.amulet.utils import (
    AmuletUtils,
    amulet,
)


@contextmanager
def captured_output():
    """Simple context manager to capture stdout/stderr.

    Source: http://stackoverflow.com/a/17981937/56219.
    """
    new_out, new_err = six.StringIO(), six.StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_out, old_err


class FakeSentry(object):

    def __init__(self, name="foo"):
        self.commands = {}
        self.info = {"unit_name": name}

    def run(self, command):
        return self.commands[command]

    def ssh(self, command):
        return self.commands[command]

    def run_action(self, action, action_args=None):
        return 'action-id'


class ValidateServicesByNameTestCase(unittest.TestCase):

    def setUp(self):
        self.utils = AmuletUtils()
        self.sentry_unit = FakeSentry()

    def test_errors_for_unknown_upstart_service(self):
        """
        Returns a message if the Upstart service is unknown.
        """
        self.sentry_unit.commands["lsb_release -cs"] = "trusty", 0
        self.sentry_unit.commands["sudo status foo"] = (
            "status: Unknown job: foo", 1)

        result = self.utils.validate_services_by_name(
            {self.sentry_unit: ["foo"]})
        self.assertIsNotNone(result)

    def test_none_for_started_upstart_service(self):
        """
        Returns None if the Upstart service is running.
        """
        self.sentry_unit.commands["lsb_release -cs"] = "trusty", 0
        self.sentry_unit.commands["sudo status foo"] = (
            "foo start/running, process 42", 0)

        result = self.utils.validate_services_by_name(
            {self.sentry_unit: ["foo"]})
        self.assertIsNone(result)

    def test_errors_for_stopped_upstart_service(self):
        """
        Returns a message if the Upstart service is stopped.
        """
        self.sentry_unit.commands["lsb_release -cs"] = "trusty", 0
        self.sentry_unit.commands["sudo status foo"] = "foo stop/waiting", 0

        result = self.utils.validate_services_by_name(
            {self.sentry_unit: ["foo"]})
        self.assertIsNotNone(result)

    def test_errors_for_unknown_systemd_service(self):
        """
        Returns a message if a systemd service is unknown.
        """
        self.sentry_unit.commands["lsb_release -cs"] = "vivid", 0
        self.sentry_unit.commands["sudo service foo status"] = (u"""\
\u25cf foo.service
   Loaded: not-found (Reason: No such file or directory)
   Active: inactive (dead)
""", 3)

        result = self.utils.validate_services_by_name({
            self.sentry_unit: ["foo"]})
        self.assertIsNotNone(result)

    def test_none_for_started_systemd_service(self):
        """
        Returns None if a systemd service is running.
        """
        self.sentry_unit.commands["lsb_release -cs"] = "vivid", 0
        self.sentry_unit.commands["sudo service foo status"] = (u"""\
\u25cf foo.service - Foo
   Loaded: loaded (/lib/systemd/system/foo.service; enabled)
   Active: active (exited) since Thu 1970-01-01 00:00:00 UTC; 42h 42min ago
 Main PID: 3 (code=exited, status=0/SUCCESS)
   CGroup: /system.slice/foo.service
""", 0)
        result = self.utils.validate_services_by_name(
            {self.sentry_unit: ["foo"]})
        self.assertIsNone(result)

    def test_errors_for_stopped_systemd_service(self):
        """
        Returns a message if a systemd service is stopped.
        """
        self.sentry_unit.commands["lsb_release -cs"] = "vivid", 0
        self.sentry_unit.commands["sudo service foo status"] = (u"""\
\u25cf foo.service - Foo
   Loaded: loaded (/lib/systemd/system/foo.service; disabled)
   Active: inactive (dead)
""", 3)
        result = self.utils.validate_services_by_name(
            {self.sentry_unit: ["foo"]})
        self.assertIsNotNone(result)


class RunActionTestCase(unittest.TestCase):

    def setUp(self):
        self.utils = AmuletUtils()
        self.sentry_unit = FakeSentry()

    def test_returns_action_id(self):
        """Returns action_id."""

        self.assertEqual("action-id", self.utils.run_action(
            self.sentry_unit, "foo"))


class WaitActionTestCase(unittest.TestCase):

    def setUp(self):
        self.utils = AmuletUtils()

    @patch.object(amulet.actions, "get_action_output")
    def test_returns_true_if_completed(self, get_action_output):
        """JSON output is parsed and returns True if the action completed."""

        get_action_output.return_value = {"status": "completed"}

        self.assertTrue(self.utils.wait_on_action("action-id"))
        get_action_output.assert_called_with("action-id", full_output=True)

    @patch.object(amulet.actions, "get_action_output")
    def test_returns_false_if_still_running(self, get_action_output):
        """
        JSON output is parsed and returns False if the action is still running.
        """
        get_action_output.return_value = {"status": "running"}

        self.assertFalse(self.utils.wait_on_action("action-id"))
        get_action_output.assert_called_with("action-id", full_output=True)

    @patch.object(amulet.actions, "get_action_output")
    def test_returns_false_if_no_status(self, get_action_output):
        """
        JSON output is parsed and returns False if there is no action status.
        """
        get_action_output.return_value = {}

        self.assertFalse(self.utils.wait_on_action("action-id"))
        get_action_output.assert_called_with("action-id", full_output=True)


class GetProcessIdListTestCase(unittest.TestCase):

    def setUp(self):
        self.utils = AmuletUtils()
        self.sentry_unit = FakeSentry()

    def test_returns_pids(self):
        """
        Normal execution returns a list of pids
        """
        self.sentry_unit.commands['pidof -x "foo"'] = ("123 124 125", 0)
        result = self.utils.get_process_id_list(self.sentry_unit, "foo")
        self.assertEqual(["123", "124", "125"], result)

    def test_fails_if_no_process_found(self):
        """
        By default, the expectation is that a process is running. Failure
        to find a given process results in an amulet.FAIL being
        raised.
        """
        self.sentry_unit.commands['pidof -x "foo"'] = ("", 1)
        with self.assertRaises(SystemExit) as cm, captured_output() as (
                out, err):
            self.utils.get_process_id_list(self.sentry_unit, "foo")
        the_exception = cm.exception
        self.assertEqual(1, the_exception.code)
        self.assertEqual(
            'foo `pidof -x "foo"` returned 1', out.getvalue().rstrip())

    def test_looks_for_scripts(self):
        """
        pidof command uses -x to return a list of pids of scripts
        """
        self.sentry_unit.commands["pidof foo"] = ("", 1)
        self.sentry_unit.commands['pidof -x "foo"'] = ("123 124 125", 0)
        result = self.utils.get_process_id_list(self.sentry_unit, "foo")
        self.assertEqual(["123", "124", "125"], result)

    def test_expect_no_pid(self):
        """
        By setting expectation that there are no pids running the logic
        about when to fail is reversed.
        """
        self.sentry_unit.commands[
            'pidof -x "foo" || exit 0 && exit 1'] = ("", 0)
        self.sentry_unit.commands[
            'pidof -x "bar" || exit 0 && exit 1'] = ("", 1)
        result = self.utils.get_process_id_list(
            self.sentry_unit, "foo", expect_success=False)
        self.assertEqual([], result)
        with self.assertRaises(SystemExit) as cm, captured_output() as (
                out, err):
            self.utils.get_process_id_list(
                self.sentry_unit, "bar", expect_success=False)
        the_exception = cm.exception
        self.assertEqual(1, the_exception.code)
        self.assertEqual(
            'foo `pidof -x "bar" || exit 0 && exit 1` returned 1',
            out.getvalue().rstrip())


class GetUnitProcessIdsTestCase(unittest.TestCase):

    def setUp(self):
        self.utils = AmuletUtils()
        self.sentry_unit = FakeSentry()

    def test_returns_map(self):
        """
        Normal execution returns a dictionary mapping process names to
        PIDs for each unit.
        """
        second_sentry = FakeSentry(name="bar")
        self.sentry_unit.commands['pidof -x "foo"'] = ("123 124", 0)
        second_sentry.commands['pidof -x "bar"'] = ("456 457", 0)

        result = self.utils.get_unit_process_ids({
            self.sentry_unit: ["foo"], second_sentry: ["bar"]})
        self.assertEqual({
            self.sentry_unit: {"foo": ["123", "124"]},
            second_sentry: {"bar": ["456", "457"]}}, result)

    def test_expect_failure(self):
        """
        Expected failures return empty lists.
        """
        second_sentry = FakeSentry(name="bar")
        self.sentry_unit.commands[
            'pidof -x "foo" || exit 0 && exit 1'] = ("", 0)
        second_sentry.commands['pidof -x "bar" || exit 0 && exit 1'] = ("", 0)

        result = self.utils.get_unit_process_ids(
            {self.sentry_unit: ["foo"], second_sentry: ["bar"]},
            expect_success=False)
        self.assertEqual({
            self.sentry_unit: {"foo": []},
            second_sentry: {"bar": []}}, result)


class StatusGetTestCase(unittest.TestCase):

    def setUp(self):
        self.utils = AmuletUtils()
        self.sentry_unit = FakeSentry()

    def test_status_get(self):
        """
        We can get the status of a unit.
        """
        self.sentry_unit.commands[
            "status-get --format=json --include-data"] = (
                """{"status": "active", "message": "foo"}""", 0)
        self.assertEqual(self.utils.status_get(self.sentry_unit),
                         (u"active", u"foo"))

    def test_status_get_missing_command(self):
        """
        Older releases of Juju have no status-get command.  In those
        cases we should return the "unknown" status.
        """
        self.sentry_unit.commands[
            "status-get --format=json --include-data"] = (
                "status-get: command not found", 127)
        self.assertEqual(self.utils.status_get(self.sentry_unit),
                         (u"unknown", u""))


class ValidateServicesByProcessIDTestCase(unittest.TestCase):

    def setUp(self):
        self.utils = AmuletUtils()
        self.sentry_unit = FakeSentry()

    def test_accepts_list_wrong(self):
        """
        Validates that it can accept a list
        """
        expected = {self.sentry_unit: {"foo": [3, 4]}}
        actual = {self.sentry_unit: {"foo": [12345, 67890]}}
        result = self.utils.validate_unit_process_ids(expected, actual)
        self.assertIsNotNone(result)

    def test_accepts_list(self):
        """
        Validates that it can accept a list
        """
        expected = {self.sentry_unit: {"foo": [2, 3]}}
        actual = {self.sentry_unit: {"foo": [12345, 67890]}}
        result = self.utils.validate_unit_process_ids(expected, actual)
        self.assertIsNone(result)

    def test_accepts_string(self):
        """
        Validates that it can accept a string
        """
        expected = {self.sentry_unit: {"foo": 2}}
        actual = {self.sentry_unit: {"foo": [12345, 67890]}}
        result = self.utils.validate_unit_process_ids(expected, actual)
        self.assertIsNone(result)

    def test_accepts_string_wrong(self):
        """
        Validates that it can accept a string
        """
        expected = {self.sentry_unit: {"foo": 3}}
        actual = {self.sentry_unit: {"foo": [12345, 67890]}}
        result = self.utils.validate_unit_process_ids(expected, actual)
        self.assertIsNotNone(result)

    def test_accepts_bool(self):
        """
        Validates that it can accept a boolean
        """
        expected = {self.sentry_unit: {"foo": True}}
        actual = {self.sentry_unit: {"foo": [12345, 67890]}}
        result = self.utils.validate_unit_process_ids(expected, actual)
        self.assertIsNone(result)

    def test_accepts_bool_wrong(self):
        """
        Validates that it can accept a boolean
        """
        expected = {self.sentry_unit: {"foo": True}}
        actual = {self.sentry_unit: {"foo": []}}
        result = self.utils.validate_unit_process_ids(expected, actual)
        self.assertIsNotNone(result)



# Copyright 2013 Canonical Ltd.
#
# Authors:
#  Charm Helpers Developers <juju@lists.ubuntu.com>
import mock
import os
import shutil
import tempfile
import unittest
import yaml

import six

import charmhelpers.contrib.templating.contexts


class JujuState2YamlTestCase(unittest.TestCase):
    maxDiff = None

    unit_data = {
        'private-address': '10.0.3.2',
        'public-address': '123.123.123.123',
    }

    def setUp(self):
        super(JujuState2YamlTestCase, self).setUp()

        # Hookenv patches (a single patch to hookenv doesn't work):
        patcher = mock.patch('charmhelpers.core.hookenv.config')
        self.mock_config = patcher.start()
        self.addCleanup(patcher.stop)
        patcher = mock.patch('charmhelpers.core.hookenv.relation_get')
        self.mock_relation_get = patcher.start()
        self.mock_relation_get.return_value = {}
        self.addCleanup(patcher.stop)
        patcher = mock.patch('charmhelpers.core.hookenv.relations')
        self.mock_relations = patcher.start()
        self.mock_relations.return_value = {
            'wsgi-file': {},
            'website': {},
            'nrpe-external-master': {},
        }
        self.addCleanup(patcher.stop)
        patcher = mock.patch('charmhelpers.core.hookenv.relation_type')
        self.mock_relation_type = patcher.start()
        self.mock_relation_type.return_value = None
        self.addCleanup(patcher.stop)
        patcher = mock.patch('charmhelpers.core.hookenv.local_unit')
        self.mock_local_unit = patcher.start()
        self.addCleanup(patcher.stop)
        patcher = mock.patch('charmhelpers.core.hookenv.relations_of_type')
        self.mock_relations_of_type = patcher.start()
        self.addCleanup(patcher.stop)
        self.mock_relations_of_type.return_value = []

        def unit_get_data(argument):
            "dummy unit_get that accesses dummy unit data"
            return self.unit_data[argument]

        patcher = mock.patch(
            'charmhelpers.core.hookenv.unit_get', unit_get_data)
        self.mock_unit_get = patcher.start()
        self.addCleanup(patcher.stop)

        # patches specific to this test class.
        etc_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, etc_dir)
        self.context_path = os.path.join(etc_dir, 'some', 'context')

        patcher = mock.patch.object(charmhelpers.contrib.templating.contexts,
                                    'charm_dir', '/tmp/charm_dir')
        patcher.start()
        self.addCleanup(patcher.stop)

    def default_context(self):
        return {
            "charm_dir": "/tmp/charm_dir",
            "group_code_owner": "webops_deploy",
            "user_code_runner": "ubunet",
            "current_relation": {},
            "relations_full": {
                'wsgi-file': {},
                'website': {},
                'nrpe-external-master': {},
            },
            "relations": {
                'wsgi-file': [],
                'website': [],
                'nrpe-external-master': [],
            },
            "local_unit": "click-index/3",
            "unit_private_address": "10.0.3.2",
            "unit_public_address": "123.123.123.123",
        }

    def test_output_with_empty_relation(self):
        self.mock_config.return_value = {
            'group_code_owner': 'webops_deploy',
            'user_code_runner': 'ubunet',
        }
        self.mock_local_unit.return_value = "click-index/3"

        charmhelpers.contrib.templating.contexts.juju_state_to_yaml(
            self.context_path)

        with open(self.context_path, 'r') as context_file:
            result = yaml.safe_load(context_file.read())
            expected = self.default_context()
            self.assertEqual(expected, result)

    def test_output_with_no_relation(self):
        self.mock_config.return_value = {
            'group_code_owner': 'webops_deploy',
            'user_code_runner': 'ubunet',
        }
        self.mock_local_unit.return_value = "click-index/3"
        self.mock_relation_get.return_value = None

        charmhelpers.contrib.templating.contexts.juju_state_to_yaml(
            self.context_path)

        with open(self.context_path, 'r') as context_file:
            result = yaml.safe_load(context_file.read())
            expected = self.default_context()
            self.assertEqual(expected, result)

    def test_output_with_relation(self):
        self.mock_config.return_value = {
            'group_code_owner': 'webops_deploy',
            'user_code_runner': 'ubunet',
        }
        self.mock_relation_type.return_value = 'wsgi-file'
        self.mock_relation_get.return_value = {
            'relation_key1': 'relation_value1',
            'relation_key2': 'relation_value2',
        }
        self.mock_relations.return_value = {
            'wsgi-file': {
                six.u('wsgi-file:0'): {
                    six.u('gunicorn/1'): {
                        six.u('private-address'): six.u('10.0.3.99'),
                    },
                    'click-index/3': {
                        six.u('wsgi_group'): six.u('ubunet'),
                    },
                },
            },
            'website': {},
            'nrpe-external-master': {},
        }
        self.mock_local_unit.return_value = "click-index/3"

        charmhelpers.contrib.templating.contexts.juju_state_to_yaml(
            self.context_path)

        with open(self.context_path, 'r') as context_file:
            result = yaml.safe_load(context_file.read())
            expected = self.default_context()
            expected['current_relation'] = {
                "relation_key1": "relation_value1",
                "relation_key2": "relation_value2",
            }
            expected["wsgi_file:relation_key1"] = "relation_value1"
            expected["wsgi_file:relation_key2"] = "relation_value2"
            expected["relations_full"]['wsgi-file'] = {
                'wsgi-file:0': {
                    'gunicorn/1': {
                        six.u('private-address'): six.u('10.0.3.99')},
                    'click-index/3': {six.u('wsgi_group'): six.u('ubunet')},
                },
            }
            expected["relations"]["wsgi-file"] = [
                {
                    '__relid__': 'wsgi-file:0',
                    '__unit__': 'gunicorn/1',
                    'private-address': '10.0.3.99',
                }
            ]
            self.assertEqual(expected, result)

    def test_relation_with_separator(self):
        self.mock_config.return_value = {
            'group_code_owner': 'webops_deploy',
            'user_code_runner': 'ubunet',
        }
        self.mock_relation_type.return_value = 'wsgi-file'
        self.mock_relation_get.return_value = {
            'relation_key1': 'relation_value1',
            'relation_key2': 'relation_value2',
        }
        self.mock_local_unit.return_value = "click-index/3"

        charmhelpers.contrib.templating.contexts.juju_state_to_yaml(
            self.context_path, namespace_separator='__')

        with open(self.context_path, 'r') as context_file:
            result = yaml.safe_load(context_file.read())
            expected = self.default_context()
            expected['current_relation'] = {
                "relation_key1": "relation_value1",
                "relation_key2": "relation_value2",
            }
            expected["wsgi_file__relation_key1"] = "relation_value1"
            expected["wsgi_file__relation_key2"] = "relation_value2"
            self.assertEqual(expected, result)

    def test_keys_with_hyphens(self):
        self.mock_config.return_value = {
            'group_code_owner': 'webops_deploy',
            'user_code_runner': 'ubunet',
            'private-address': '10.1.1.10',
        }
        self.mock_local_unit.return_value = "click-index/3"
        self.mock_relation_get.return_value = None

        charmhelpers.contrib.templating.contexts.juju_state_to_yaml(
            self.context_path)

        with open(self.context_path, 'r') as context_file:
            result = yaml.safe_load(context_file.read())
            expected = self.default_context()
            expected["private-address"] = "10.1.1.10"
            self.assertEqual(expected, result)

    def test_keys_with_hypens_not_allowed_in_keys(self):
        self.mock_config.return_value = {
            'group_code_owner': 'webops_deploy',
            'user_code_runner': 'ubunet',
            'private-address': '10.1.1.10',
        }
        self.mock_local_unit.return_value = "click-index/3"
        self.mock_relation_type.return_value = 'wsgi-file'
        self.mock_relation_get.return_value = {
            'relation-key1': 'relation_value1',
            'relation-key2': 'relation_value2',
        }

        charmhelpers.contrib.templating.contexts.juju_state_to_yaml(
            self.context_path, allow_hyphens_in_keys=False,
            namespace_separator='__')

        with open(self.context_path, 'r') as context_file:
            result = yaml.safe_load(context_file.read())
            expected = self.default_context()
            expected["private_address"] = "10.1.1.10"
            expected["wsgi_file__relation_key1"] = "relation_value1"
            expected["wsgi_file__relation_key2"] = "relation_value2"
            expected['current_relation'] = {
                "relation-key1": "relation_value1",
                "relation-key2": "relation_value2",
            }
            self.assertEqual(expected, result)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from charmhelpers.core.sysctl import create
import io
from mock import patch, MagicMock
from subprocess import CalledProcessError
import unittest
import tempfile

import six
if not six.PY3:
    builtin_open = '__builtin__.open'
else:
    builtin_open = 'builtins.open'

__author__ = 'Jorge Niedbalski R. <jorge.niedbalski@canonical.com>'


TO_PATCH = [
    'log',
    'check_call',
    'is_container',
]


class SysctlTests(unittest.TestCase):
    def setUp(self):
        self.tempfile = tempfile.NamedTemporaryFile(delete=False)
        for m in TO_PATCH:
            setattr(self, m, self._patch(m))

    def _patch(self, method):
        _m = patch('charmhelpers.core.sysctl.' + method)
        mock = _m.start()
        self.addCleanup(_m.stop)
        return mock

    @patch(builtin_open)
    def test_create(self, mock_open):
        """Test create sysctl method"""
        _file = MagicMock(spec=io.FileIO)
        mock_open.return_value = _file

        create('{"kernel.max_pid": 1337}', "/etc/sysctl.d/test-sysctl.conf")

        _file.__enter__().write.assert_called_with("kernel.max_pid=1337\n")

        self.log.assert_called_with(
            "Updating sysctl_file: /etc/sysctl.d/test-sysctl.conf"
            " values: {'kernel.max_pid': 1337}",
            level='DEBUG')

        self.check_call.assert_called_with([
            "sysctl", "-p",
            "/etc/sysctl.d/test-sysctl.conf"])

    @patch(builtin_open)
    def test_create_with_ignore(self, mock_open):
        """Test create sysctl method"""
        _file = MagicMock(spec=io.FileIO)
        mock_open.return_value = _file

        create('{"kernel.max_pid": 1337}',
               "/etc/sysctl.d/test-sysctl.conf",
               ignore=True)

        _file.__enter__().write.assert_called_with("kernel.max_pid=1337\n")

        self.log.assert_called_with(
            "Updating sysctl_file: /etc/sysctl.d/test-sysctl.conf"
            " values: {'kernel.max_pid': 1337}",
            level='DEBUG')

        self.check_call.assert_called_with([
            "sysctl", "-p",
            "/etc/sysctl.d/test-sysctl.conf", "-e"])

    @patch(builtin_open)
    def test_create_with_dict(self, mock_open):
        """Test create sysctl method"""
        _file = MagicMock(spec=io.FileIO)
        mock_open.return_value = _file

        create({"kernel.max_pid": 1337}, "/etc/sysctl.d/test-sysctl.conf")

        _file.__enter__().write.assert_called_with("kernel.max_pid=1337\n")

        self.log.assert_called_with(
            "Updating sysctl_file: /etc/sysctl.d/test-sysctl.conf"
            " values: {'kernel.max_pid': 1337}",
            level='DEBUG')

        self.check_call.assert_called_with([
            "sysctl", "-p",
            "/etc/sysctl.d/test-sysctl.conf"])

    @patch(builtin_open)
    def test_create_invalid_argument(self, mock_open):
        """Test create sysctl with an invalid argument"""
        _file = MagicMock(spec=io.FileIO)
        mock_open.return_value = _file

        create('{"kernel.max_pid": 1337 xxxx', "/etc/sysctl.d/test-sysctl.conf")

        self.log.assert_called_with(
            'Error parsing YAML sysctl_dict: {"kernel.max_pid": 1337 xxxx',
            level='ERROR')

    @patch(builtin_open)
    def test_create_raises(self, mock_open):
        """CalledProcessErrors are propagated for non-container machines."""
        _file = MagicMock(spec=io.FileIO)
        mock_open.return_value = _file

        self.is_container.return_value = False
        self.check_call.side_effect = CalledProcessError(1, 'sysctl')

        with self.assertRaises(CalledProcessError):
            create('{"kernel.max_pid": 1337}', "/etc/sysctl.d/test-sysctl.conf")

    @patch(builtin_open)
    def test_create_raises_container(self, mock_open):
        """CalledProcessErrors are logged for containers."""
        _file = MagicMock(spec=io.FileIO)
        mock_open.return_value = _file

        self.is_container.return_value = True
        self.check_call.side_effect = CalledProcessError(1, 'sysctl', 'foo')

        create('{"kernel.max_pid": 1337}', "/etc/sysctl.d/test-sysctl.conf")
        self.log.assert_called_with(
            'Error setting some sysctl keys in this container: foo',
            level='WARNING')



import subprocess
import unittest

from mock import patch, call, MagicMock

import charmhelpers.contrib.network.ovs as ovs

from tests.helpers import patch_open


# NOTE(fnordahl): some functions drectly under the ``contrib.network.ovs``
# module have their unit tests in the ``test_ovs.py`` module in the
# ``tests.contrib.network.ovs`` package.


GOOD_CERT = '''Certificate:
    Data:
        Version: 1 (0x0)
        Serial Number: 13798680962510501282 (0xbf7ec33a136235a2)
    Signature Algorithm: sha1WithRSAEncryption
        Issuer: C=US, ST=CA, L=Palo Alto, O=Open vSwitch, OU=Open vSwitch
        Validity
            Not Before: Jun 28 17:02:19 2013 GMT
            Not After : Jun 28 17:02:19 2019 GMT
        Subject: C=US, ST=CA, L=Palo Alto, O=Open vSwitch, OU=Open vSwitch
        Subject Public Key Info:
            Public Key Algorithm: rsaEncryption
                Public-Key: (2048 bit)
                Modulus:
                    00:e8:a7:db:0a:6d:c0:16:4a:14:96:1d:74:91:15:
                    64:3f:ae:2a:54:be:2a:fe:10:14:9a:73:39:d8:58:
                    74:7f:ab:d5:f2:39:aa:9a:27:7c:31:82:f8:74:42:
                    46:8d:c5:3b:42:55:52:be:75:7f:a5:b1:ec:d5:29:
                    9f:62:0e:de:31:27:2b:95:1f:24:0d:ca:8c:48:30:
                    96:9f:ba:b7:9d:eb:c1:bd:93:05:e3:d8:ca:66:5a:
                    e9:cb:a5:7a:3a:8d:27:e2:05:9d:88:fc:a9:ef:af:
                    47:4c:66:ce:c6:43:73:1a:85:f4:5f:b9:53:5b:29:
                    f3:c3:23:1f:0c:20:95:11:50:71:b2:f6:01:23:3f:
                    66:0f:5c:43:c2:90:fb:e5:98:73:98:e9:38:bb:1f:
                    1b:89:97:1e:dc:d7:98:07:68:32:ec:da:1d:69:0b:
                    e2:df:40:fb:64:52:e5:e9:40:27:b0:ca:73:21:51:
                    f6:8f:00:20:c0:2b:1a:d4:01:c2:32:38:9d:d1:8d:
                    88:71:46:a9:42:0d:ee:3b:1c:88:db:27:69:49:f9:
                    60:34:70:61:3d:60:df:7e:e4:e1:1d:c6:16:89:05:
                    ba:31:06:eb:88:b5:78:94:5d:8c:9d:88:fe:f2:c2:
                    80:a1:04:15:d3:84:85:d3:aa:5a:1d:53:5c:f8:57:
                    ae:61
                Exponent: 65537 (0x10001)
    Signature Algorithm: sha1WithRSAEncryption
         14:7e:ca:c3:fc:93:60:9f:80:e0:65:2e:ef:41:2d:f9:af:77:
         da:6d:e2:e0:11:70:17:fb:e5:67:4c:f0:ad:39:ec:96:ef:fe:
         d5:95:94:70:e5:52:31:68:63:8c:ea:b3:a1:8e:02:e2:91:4b:
         a8:8c:07:86:fd:80:98:a2:b1:90:2b:9c:2e:ab:f4:73:9d:8f:
         fd:31:b9:8f:fe:6c:af:d6:bf:72:44:89:08:93:19:ef:2b:c3:
         7c:ab:ba:bc:57:ca:f1:17:e4:e8:81:40:ca:65:df:84:be:10:
         2c:42:46:af:d2:e0:0d:df:5d:56:53:65:13:e0:20:55:b4:ee:
         cd:5e:b5:c4:97:1d:3e:a6:c1:9c:7e:b8:87:ee:64:78:a5:59:
         e5:b2:79:47:9a:8e:59:fa:c4:18:ea:27:fd:a2:d5:76:d0:ae:
         d9:05:f6:0e:23:ca:7d:66:a1:ba:18:67:f5:6d:bb:51:5a:f5:
         52:e9:17:bb:63:15:24:b4:61:25:9f:d9:9c:89:58:93:9a:c3:
         74:55:72:3e:f9:ff:ef:54:7d:e8:28:78:ba:3c:c7:15:ba:b9:
         c6:e3:8c:61:cb:a9:ed:8d:07:16:0d:8d:f6:1c:36:11:69:08:
         b8:45:7d:fc:fd:d1:ab:2d:9b:4e:9c:dd:11:78:50:c7:87:9f:
         4a:24:9c:a0
-----BEGIN CERTIFICATE-----
MIIDwjCCAqoCCQC/fsM6E2I1ojANBgkqhkiG9w0BAQUFADCBojELMAkGA1UEBhMC
VVMxCzAJBgNVBAgTAkNBMRIwEAYDVQQHEwlQYWxvIEFsdG8xFTATBgNVBAoTDE9w
ZW4gdlN3aXRjaDEfMB0GA1UECxMWT3BlbiB2U3dpdGNoIGNlcnRpZmllcjE6MDgG
A1UEAxMxb3ZzY2xpZW50IGlkOjU4MTQ5N2E1LWJjMDAtNGVjYy1iNzkwLTU3NTZj
ZWUxNmE0ODAeFw0xMzA2MjgxNzAyMTlaFw0xOTA2MjgxNzAyMTlaMIGiMQswCQYD
VQQGEwJVUzELMAkGA1UECBMCQ0ExEjAQBgNVBAcTCVBhbG8gQWx0bzEVMBMGA1UE
ChMMT3BlbiB2U3dpdGNoMR8wHQYDVQQLExZPcGVuIHZTd2l0Y2ggY2VydGlmaWVy
MTowOAYDVQQDEzFvdnNjbGllbnQgaWQ6NTgxNDk3YTUtYmMwMC00ZWNjLWI3OTAt
NTc1NmNlZTE2YTQ4MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA6Kfb
Cm3AFkoUlh10kRVkP64qVL4q/hAUmnM52Fh0f6vV8jmqmid8MYL4dEJGjcU7QlVS
vnV/pbHs1SmfYg7eMScrlR8kDcqMSDCWn7q3nevBvZMF49jKZlrpy6V6Oo0n4gWd
iPyp769HTGbOxkNzGoX0X7lTWynzwyMfDCCVEVBxsvYBIz9mD1xDwpD75ZhzmOk4
ux8biZce3NeYB2gy7NodaQvi30D7ZFLl6UAnsMpzIVH2jwAgwCsa1AHCMjid0Y2I
cUapQg3uOxyI2ydpSflgNHBhPWDffuThHcYWiQW6MQbriLV4lF2MnYj+8sKAoQQV
04SF06paHVNc+FeuYQIDAQABMA0GCSqGSIb3DQEBBQUAA4IBAQAUfsrD/JNgn4Dg
ZS7vQS35r3fabeLgEXAX++VnTPCtOeyW7/7VlZRw5VIxaGOM6rOhjgLikUuojAeG
/YCYorGQK5wuq/RznY/9MbmP/myv1r9yRIkIkxnvK8N8q7q8V8rxF+TogUDKZd+E
vhAsQkav0uAN311WU2UT4CBVtO7NXrXElx0+psGcfriH7mR4pVnlsnlHmo5Z+sQY
6if9otV20K7ZBfYOI8p9ZqG6GGf1bbtRWvVS6Re7YxUktGEln9mciViTmsN0VXI+
+f/vVH3oKHi6PMcVurnG44xhy6ntjQcWDY32HDYRaQi4RX38/dGrLZtOnN0ReFDH
h59KJJyg
-----END CERTIFICATE-----
'''

PEM_ENCODED = '''-----BEGIN CERTIFICATE-----
MIIDwjCCAqoCCQC/fsM6E2I1ojANBgkqhkiG9w0BAQUFADCBojELMAkGA1UEBhMC
VVMxCzAJBgNVBAgTAkNBMRIwEAYDVQQHEwlQYWxvIEFsdG8xFTATBgNVBAoTDE9w
ZW4gdlN3aXRjaDEfMB0GA1UECxMWT3BlbiB2U3dpdGNoIGNlcnRpZmllcjE6MDgG
A1UEAxMxb3ZzY2xpZW50IGlkOjU4MTQ5N2E1LWJjMDAtNGVjYy1iNzkwLTU3NTZj
ZWUxNmE0ODAeFw0xMzA2MjgxNzAyMTlaFw0xOTA2MjgxNzAyMTlaMIGiMQswCQYD
VQQGEwJVUzELMAkGA1UECBMCQ0ExEjAQBgNVBAcTCVBhbG8gQWx0bzEVMBMGA1UE
ChMMT3BlbiB2U3dpdGNoMR8wHQYDVQQLExZPcGVuIHZTd2l0Y2ggY2VydGlmaWVy
MTowOAYDVQQDEzFvdnNjbGllbnQgaWQ6NTgxNDk3YTUtYmMwMC00ZWNjLWI3OTAt
NTc1NmNlZTE2YTQ4MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEA6Kfb
Cm3AFkoUlh10kRVkP64qVL4q/hAUmnM52Fh0f6vV8jmqmid8MYL4dEJGjcU7QlVS
vnV/pbHs1SmfYg7eMScrlR8kDcqMSDCWn7q3nevBvZMF49jKZlrpy6V6Oo0n4gWd
iPyp769HTGbOxkNzGoX0X7lTWynzwyMfDCCVEVBxsvYBIz9mD1xDwpD75ZhzmOk4
ux8biZce3NeYB2gy7NodaQvi30D7ZFLl6UAnsMpzIVH2jwAgwCsa1AHCMjid0Y2I
cUapQg3uOxyI2ydpSflgNHBhPWDffuThHcYWiQW6MQbriLV4lF2MnYj+8sKAoQQV
04SF06paHVNc+FeuYQIDAQABMA0GCSqGSIb3DQEBBQUAA4IBAQAUfsrD/JNgn4Dg
ZS7vQS35r3fabeLgEXAX++VnTPCtOeyW7/7VlZRw5VIxaGOM6rOhjgLikUuojAeG
/YCYorGQK5wuq/RznY/9MbmP/myv1r9yRIkIkxnvK8N8q7q8V8rxF+TogUDKZd+E
vhAsQkav0uAN311WU2UT4CBVtO7NXrXElx0+psGcfriH7mR4pVnlsnlHmo5Z+sQY
6if9otV20K7ZBfYOI8p9ZqG6GGf1bbtRWvVS6Re7YxUktGEln9mciViTmsN0VXI+
+f/vVH3oKHi6PMcVurnG44xhy6ntjQcWDY32HDYRaQi4RX38/dGrLZtOnN0ReFDH
h59KJJyg
-----END CERTIFICATE-----'''

BAD_CERT = ''' NO MARKERS '''
TO_PATCH = [
    "apt_install",
    "log",
    "hashlib",
]


class OVSHelpersTest(unittest.TestCase):

    def setUp(self):
        for m in TO_PATCH:
            setattr(self, m, self._patch(m))

    def _patch(self, method):
        _m = patch('charmhelpers.contrib.network.ovs.' + method)
        mock = _m.start()
        self.addCleanup(_m.stop)
        return mock

    @patch('subprocess.check_output')
    def test_get_bridges(self, check_output):
        check_output.return_value = b"br1\n br2  "
        self.assertEqual(ovs.get_bridges(), ['br1', 'br2'])
        check_output.assert_called_once_with(['ovs-vsctl', 'list-br'])

    @patch('subprocess.check_output')
    def test_get_bridge_ports(self, check_output):
        check_output.return_value = b"p1\n p2  \np3"
        self.assertEqual(ovs.get_bridge_ports('br1'), ['p1', 'p2', 'p3'])
        check_output.assert_called_once_with(
            ['ovs-vsctl', '--', 'list-ports', 'br1'])

    @patch.object(ovs, 'get_bridges')
    @patch.object(ovs, 'get_bridge_ports')
    def test_get_bridges_and_ports_map(self, get_bridge_ports, get_bridges):
        get_bridges.return_value = ['br1', 'br2']
        get_bridge_ports.side_effect = [
            ['p1', 'p2'],
            ['p3']]
        self.assertEqual(ovs.get_bridges_and_ports_map(), {
            'br1': ['p1', 'p2'],
            'br2': ['p3'],
        })

    @patch('subprocess.check_call')
    def test_del_bridge(self, check_call):
        ovs.del_bridge('test')
        check_call.assert_called_with(["ovs-vsctl", "--", "--if-exists",
                                       "del-br", 'test'])
        self.assertTrue(self.log.call_count == 1)

    @patch('subprocess.check_call')
    def test_del_bridge_port(self, check_call):
        ovs.del_bridge_port('test', 'eth1')
        check_call.assert_has_calls([
            call(["ovs-vsctl", "--", "--if-exists", "del-port",
                  'test', 'eth1']),
            call(['ip', 'link', 'set', 'eth1', 'down']),
            call(['ip', 'link', 'set', 'eth1', 'promisc', 'off'])
        ])
        self.assertTrue(self.log.call_count == 1)

    @patch.object(ovs, 'port_to_br')
    @patch.object(ovs, 'add_bridge_port')
    @patch('subprocess.check_call')
    def test_add_ovsbridge_linuxbridge(self, check_call,
                                       add_bridge_port,
                                       port_to_br):
        port_to_br.return_value = None
        with patch_open() as (mock_open, mock_file):
            ovs.add_ovsbridge_linuxbridge('br-ex', 'br-eno1', ifdata={
                'external-ids': {'mycharm': 'br-ex'}
            })

        check_call.assert_called_with(['ifup', 'veth-br-eno1'])
        add_bridge_port.assert_called_with(
            'br-ex', 'veth-br-eno1', ifdata={
                'external-ids': {'mycharm': 'br-ex'}
            }
        )

    @patch.object(ovs, 'port_to_br')
    @patch.object(ovs, 'add_bridge_port')
    @patch('subprocess.check_call')
    def test_add_ovsbridge_linuxbridge_already_direct_wired(self,
                                                            check_call,
                                                            add_bridge_port,
                                                            port_to_br):
        port_to_br.return_value = 'br-ex'
        ovs.add_ovsbridge_linuxbridge('br-ex', 'br-eno1')
        check_call.assert_not_called()
        add_bridge_port.assert_not_called()

    @patch.object(ovs, 'port_to_br')
    @patch.object(ovs, 'add_bridge_port')
    @patch('subprocess.check_call')
    def test_add_ovsbridge_linuxbridge_longname(self, check_call,
                                                add_bridge_port,
                                                port_to_br):
        port_to_br.return_value = None
        mock_hasher = MagicMock()
        mock_hasher.hexdigest.return_value = '12345678901234578910'
        self.hashlib.sha256.return_value = mock_hasher
        with patch_open() as (mock_open, mock_file):
            ovs.add_ovsbridge_linuxbridge('br-ex', 'br-reallylongname')

        check_call.assert_called_with(['ifup', 'cvb12345678-10'])
        add_bridge_port.assert_called_with(
            'br-ex', 'cvb12345678-10', ifdata=None
        )

    @patch('os.path.exists')
    def test_is_linuxbridge_interface_false(self, exists):
        exists.return_value = False
        result = ovs.is_linuxbridge_interface('eno1')
        self.assertFalse(result)

    @patch('os.path.exists')
    def test_is_linuxbridge_interface_true(self, exists):
        exists.return_value = True
        result = ovs.is_linuxbridge_interface('eno1')
        self.assertTrue(result)

    @patch('subprocess.check_call')
    def test_set_manager(self, check_call):
        ovs.set_manager('manager')
        check_call.assert_called_with(['ovs-vsctl', 'set-manager',
                                       'ssl:manager'])
        self.assertTrue(self.log.call_count == 1)

    @patch('subprocess.check_call')
    def test_set_Open_vSwitch_column_value(self, check_call):
        ovs.set_Open_vSwitch_column_value('other_config:foo=bar')
        check_call.assert_called_with(['ovs-vsctl', 'set',
                                       'Open_vSwitch', '.', 'other_config:foo=bar'])
        self.assertTrue(self.log.call_count == 1)

    @patch('os.path.exists')
    def test_get_certificate_good_cert(self, exists):
        exists.return_value = True
        with patch_open() as (mock_open, mock_file):
            mock_file.read.return_value = GOOD_CERT
            self.assertEqual(ovs.get_certificate(), PEM_ENCODED)
        self.assertTrue(self.log.call_count == 1)

    @patch('os.path.exists')
    def test_get_certificate_bad_cert(self, exists):
        exists.return_value = True
        with patch_open() as (mock_open, mock_file):
            mock_file.read.return_value = BAD_CERT
            self.assertRaises(RuntimeError, ovs.get_certificate)
        self.assertTrue(self.log.call_count == 1)

    @patch('os.path.exists')
    def test_get_certificate_missing(self, exists):
        exists.return_value = False
        self.assertIsNone(ovs.get_certificate())
        self.assertTrue(self.log.call_count == 1)

    @patch('os.path.exists')
    @patch.object(ovs, 'service')
    def test_full_restart(self, service, exists):
        exists.return_value = False
        ovs.full_restart()
        service.assert_called_with('force-reload-kmod', 'openvswitch-switch')

    @patch('os.path.exists')
    @patch.object(ovs, 'service')
    def test_full_restart_upstart(self, service, exists):
        exists.return_value = True
        ovs.full_restart()
        service.assert_called_with('start', 'openvswitch-force-reload-kmod')

    @patch('subprocess.check_output')
    def test_port_to_br(self, check_output):
        check_output.return_value = b'br-ex'
        self.assertEqual(ovs.port_to_br('br-lb'),
                         'br-ex')

    @patch('subprocess.check_output')
    def test_port_to_br_not_found(self, check_output):
        check_output.side_effect = subprocess.CalledProcessError(1, 'not found')
        self.assertEqual(ovs.port_to_br('br-lb'), None)

    @patch('subprocess.check_call')
    def test_enable_ipfix_defaults(self, check_call):
        ovs.enable_ipfix('br-int',
                         '10.5.0.10:4739')
        check_call.assert_called_once_with([
            'ovs-vsctl', 'set', 'Bridge', 'br-int', 'ipfix=@i', '--',
            '--id=@i', 'create', 'IPFIX',
            'targets="10.5.0.10:4739"',
            'sampling=64',
            'cache_active_timeout=60',
            'cache_max_flows=128',
        ])

    @patch('subprocess.check_call')
    def test_enable_ipfix_values(self, check_call):
        ovs.enable_ipfix('br-int',
                         '10.5.0.10:4739',
                         sampling=120,
                         cache_max_flows=24,
                         cache_active_timeout=120)
        check_call.assert_called_once_with([
            'ovs-vsctl', 'set', 'Bridge', 'br-int', 'ipfix=@i', '--',
            '--id=@i', 'create', 'IPFIX',
            'targets="10.5.0.10:4739"',
            'sampling=120',
            'cache_active_timeout=120',
            'cache_max_flows=24',
        ])

    @patch('subprocess.check_call')
    def test_disable_ipfix(self, check_call):
        ovs.disable_ipfix('br-int')
        check_call.assert_called_once_with(
            ['ovs-vsctl', 'clear', 'Bridge', 'br-int', 'ipfix']
        )



#!/usr/bin/env python
# coding: utf-8

# Copyright 2014-2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import six
import subprocess
import sys

from charmhelpers.fetch import apt_install, apt_update
from charmhelpers.core.hookenv import charm_dir, log

__author__ = "Jorge Niedbalski <jorge.niedbalski@canonical.com>"


def pip_execute(*args, **kwargs):
    """Overriden pip_execute() to stop sys.path being changed.

    The act of importing main from the pip module seems to cause add wheels
    from the /usr/share/python-wheels which are installed by various tools.
    This function ensures that sys.path remains the same after the call is
    executed.
    """
    try:
        _path = sys.path
        try:
            from pip import main as _pip_execute
        except ImportError:
            apt_update()
            if six.PY2:
                apt_install('python-pip')
            else:
                apt_install('python3-pip')
            from pip import main as _pip_execute
        _pip_execute(*args, **kwargs)
    finally:
        sys.path = _path


def parse_options(given, available):
    """Given a set of options, check if available"""
    for key, value in sorted(given.items()):
        if not value:
            continue
        if key in available:
            yield "--{0}={1}".format(key, value)


def pip_install_requirements(requirements, constraints=None, **options):
    """Install a requirements file.

    :param constraints: Path to pip constraints file.
    http://pip.readthedocs.org/en/stable/user_guide/#constraints-files
    """
    command = ["install"]

    available_options = ('proxy', 'src', 'log', )
    for option in parse_options(options, available_options):
        command.append(option)

    command.append("-r {0}".format(requirements))
    if constraints:
        command.append("-c {0}".format(constraints))
        log("Installing from file: {} with constraints {} "
            "and options: {}".format(requirements, constraints, command))
    else:
        log("Installing from file: {} with options: {}".format(requirements,
                                                               command))
    pip_execute(command)


def pip_install(package, fatal=False, upgrade=False, venv=None,
                constraints=None, **options):
    """Install a python package"""
    if venv:
        venv_python = os.path.join(venv, 'bin/pip')
        command = [venv_python, "install"]
    else:
        command = ["install"]

    available_options = ('proxy', 'src', 'log', 'index-url', )
    for option in parse_options(options, available_options):
        command.append(option)

    if upgrade:
        command.append('--upgrade')

    if constraints:
        command.extend(['-c', constraints])

    if isinstance(package, list):
        command.extend(package)
    else:
        command.append(package)

    log("Installing {} package with options: {}".format(package,
                                                        command))
    if venv:
        subprocess.check_call(command)
    else:
        pip_execute(command)


def pip_uninstall(package, **options):
    """Uninstall a python package"""
    command = ["uninstall", "-q", "-y"]

    available_options = ('proxy', 'log', )
    for option in parse_options(options, available_options):
        command.append(option)

    if isinstance(package, list):
        command.extend(package)
    else:
        command.append(package)

    log("Uninstalling {} package with options: {}".format(package,
                                                          command))
    pip_execute(command)


def pip_list():
    """Returns the list of current python installed packages
    """
    return pip_execute(["list"])


def pip_create_virtualenv(path=None):
    """Create an isolated Python environment."""
    if six.PY2:
        apt_install('python-virtualenv')
    else:
        apt_install('python3-virtualenv')

    if path:
        venv_path = path
    else:
        venv_path = os.path.join(charm_dir(), 'venv')

    if not os.path.exists(venv_path):
        subprocess.check_call(['virtualenv', venv_path])
